"""Create JSON Schema."""

#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import functools
import importlib
import inspect
import textwrap
import typing as t
from collections import defaultdict
from dataclasses import dataclass
from graphlib import TopologicalSorter
from pathlib import Path

import click
from boltons.strutils import under2camel
from numpydoc.docscrape import NumpyDocString
from toolz import dicttoolz
from tqdm.auto import tqdm

from pyxel import __version__


@dataclass
class Param:
    description: str
    annotation: t.Any


@dataclass
class ParamDefault:
    description: str
    annotation: t.Any
    default: t.Any


@dataclass
class FuncDocumentation:
    description: str
    parameters: t.Mapping[str, t.Union[Param, ParamDefault]]


@dataclass
class ModelInfo:
    model_name: str
    model_fullname: str
    model_class_name: str
    func: t.Callable


@dataclass
class ModelGroupInfo:
    name: str
    class_name: str


@dataclass(frozen=True)
class Klass:
    cls: t.Type
    base_cls: t.Optional[t.Type] = None

    @property
    def name(self) -> str:
        return self.cls.__name__


def get_documentation(func: t.Callable) -> FuncDocumentation:
    assert func.__doc__
    doc = NumpyDocString(inspect.cleandoc(func.__doc__))

    signature = inspect.signature(func)  # type: inspect.Signature

    parameters = {}

    if doc["Parameters"]:
        for params in doc["Parameters"]:
            name, *_ = params.name.split(":")
            description = "\n".join(params.desc)

            parameter = signature.parameters[name]  # type: inspect.Parameter

            if hasattr(parameter.annotation, "__name__"):
                annotation = parameter.annotation.__name__  # type: str
            else:
                annotation = str(parameter.annotation)

            if parameter.default != inspect.Parameter.empty:
                param = ParamDefault(
                    description=description,
                    annotation=annotation.replace("NoneType", "None"),
                    default=parameter.default,
                )  # type: t.Union[ParamDefault, Param]
            else:
                param = Param(description=description, annotation=annotation)

            parameters[name.strip()] = param
    else:
        for name, parameter in signature.parameters.items():
            if hasattr(parameter.annotation, "__name__"):
                annotation = parameter.annotation.__name__  # type: str
            else:
                annotation = str(parameter.annotation)

            if parameter.default != inspect.Parameter.empty:
                param = ParamDefault(
                    description="",
                    annotation=annotation.replace("NoneType", "None"),
                    default=parameter.default,
                )  # type: t.Union[ParamDefault, Param]
            else:
                param = Param(description="", annotation=annotation)

            parameters[name.strip()] = param

    return FuncDocumentation(
        description="\n".join(doc["Summary"]), parameters=parameters
    )


@functools.cache
def get_doc_from_klass(klass: Klass) -> FuncDocumentation:
    if klass.base_cls is None:
        doc = get_documentation(klass.cls)  # type: FuncDocumentation
    else:
        doc_base = get_documentation(klass.base_cls)  # type: FuncDocumentation
        doc_inherited = get_documentation(klass.cls)  # type: FuncDocumentation

        doc = FuncDocumentation(
            description=doc_inherited.description,
            parameters=dicttoolz.dissoc(doc_inherited.parameters, *doc_base.parameters),
        )

    return doc


def generate_class(klass: Klass) -> t.Iterator[str]:

    doc = get_doc_from_klass(klass)
    klass_description_lst = textwrap.wrap(doc.description)  # type: t.Sequence[str]

    yield "@schema("
    if len(klass_description_lst) == 1:
        yield f"    title={klass.name!r},"
        yield f"    description={klass_description_lst[0]!r}"
    elif len(klass_description_lst) > 1:
        yield f"    title={klass.name!r},"
        yield "    description=("
        for line in klass_description_lst:
            yield f"        {line!r}"
        yield "        )"
    else:
        yield f"    title={klass.name!r}"

    yield ")"
    yield "@dataclass"

    if klass.base_cls is None:
        yield f"class {klass.cls.__name__}:"
    else:
        yield f"class {klass.cls.__name__}({klass.base_cls.__name__}):"

    if doc.parameters:
        for name, param in doc.parameters.items():
            title = name

            yield f"    {name}: {param.annotation} = field("
            if isinstance(param, ParamDefault):
                yield f"        default={param.default!r},"
            yield "        metadata=schema("

            description_lst = textwrap.wrap(param.description)  # type: t.Sequence[str]
            if len(description_lst) == 1:
                yield f"            title={title!r},"
                yield f"            description={description_lst[0]!r}"
            elif len(description_lst) > 1:
                yield f"            title={title!r},"
                yield "            description=("
                for line in description_lst:
                    yield f"                    {line!r}"
                yield "                )"
            else:
                yield f"            title={title!r}"

            yield "        )"
            yield "    )"

    else:
        yield "    pass"

    yield ""
    yield ""


def generate_model(
    func: t.Callable,
    func_name: str,
    func_fullname: str,
    model_name: str,
) -> t.Iterator[str]:

    doc = get_documentation(func)  # type: FuncDocumentation

    yield "@schema(title='Parameters')"
    yield "@dataclass"
    yield f"class {model_name}Arguments(typing.Mapping[str, typing.Any]):"

    dct = {key: value for key, value in doc.parameters.items() if key != "detector"}

    all_defaults = all(
        [isinstance(el, ParamDefault) for el in dct.values()]
    )  # type: bool

    if dct:
        for name, param in dct.items():
            title = name

            yield f"    {name}: {param.annotation} = field("
            if isinstance(param, ParamDefault):
                yield f"        default={param.default!r},"
            yield "        metadata=schema("
            yield f"            title={title!r},"

            description_lst = textwrap.wrap(param.description)  # type: t.Sequence[str]
            if len(description_lst) == 1:
                yield f"            description={description_lst[0]!r}"
            elif len(description_lst) > 1:
                yield "            description=("
                for line in description_lst:
                    yield f"                    {line!r}"
                yield "                )"

            yield "        )"
            yield "    )"
    # else:
    #     yield "    pass"

    yield "    def __iter__(self) -> typing.Iterator[str]:"
    yield f"        return iter({tuple(dct)!r})"

    yield "    def __getitem__(self, item: typing.Any) -> typing.Any:"
    yield "        if item in tuple(self):"
    yield "            return getattr(self, item)"
    yield "        else:"
    yield "            raise KeyError"

    yield "    def __len__(self) -> int:"
    yield f"        return {len(dct)}"

    yield ""
    yield ""
    yield "@schema("
    yield f"    title=\"Model '{func_name}'\","

    description_lst = textwrap.wrap(doc.description)
    if len(description_lst) == 1:
        yield f"    description={description_lst[0]!r}"
    elif len(description_lst) > 1:
        yield "    description=("
        for line in description_lst:
            yield f"        {line!r}"
        yield "    )"

    yield ")"
    yield "@dataclass"
    yield f"class {model_name}:"
    yield "    name: str"

    if all_defaults:
        yield f"    arguments: {model_name}Arguments = field(default_factory={model_name}Arguments)"
    else:
        yield f"    arguments: {model_name}Arguments"
    yield f"    func: typing.Literal[{func_fullname!r}] = {func_fullname!r}"
    yield "    enabled: bool = True"
    yield ""
    yield ""


def get_model_info(group_name: str) -> t.Sequence[ModelInfo]:
    group_module = importlib.import_module(f"pyxel.models.{group_name}")

    lst = []  # type: t.List[ModelInfo]
    for name in dir(group_module):
        if name.startswith("__"):
            continue

        func = getattr(group_module, name)  # type: t.Callable
        if not callable(func):
            continue

        sig = inspect.signature(func)  # type: inspect.Signature

        if "detector" not in sig.parameters:
            continue

        lst.append(
            ModelInfo(
                model_name=name,
                model_fullname=f"pyxel.models.{group_name}.{name}",
                model_class_name=under2camel(f"Model_{group_name}_{name}"),
                func=func,
            )
        )

    return lst


def capitalize_title(name: str) -> str:
    return " ".join([el.capitalize() for el in name.split("_")])


def generate_group(model_groups_info: t.Sequence[ModelGroupInfo]) -> t.Iterator[str]:

    for group_info in model_groups_info:
        group_name = group_info.name

        models_info = get_model_info(group_name)  # type: t.Sequence[ModelInfo]

        for info in models_info:  # type: ModelInfo

            group_name_title = capitalize_title(group_name)
            model_title = capitalize_title(info.model_name)

            yield "#"
            yield f"# Model: {group_name_title} / {model_title}"
            yield "#"

            yield from generate_model(
                func=info.func,
                func_name=info.model_name,
                func_fullname=info.model_fullname,
                model_name=info.model_class_name,
            )

    yield "#"
    yield "# Detection pipeline"
    yield "#"
    yield "@dataclass"
    yield "class DetailedDetectionPipeline(DetectionPipeline):"

    for group_info in model_groups_info:
        group_name = group_info.name
        group_name_title = capitalize_title(group_name)

        models_info = get_model_info(group_name)
        models_class_names = [info.model_class_name for info in models_info]

        all_model_class_names = ", ".join([*models_class_names, "ModelFunction"])

        yield f"    {group_name}: typing.Sequence["
        yield "        typing.Union["
        yield f"            {all_model_class_names}"
        yield "        ]"
        yield f"    ] = field(default_factory=list, metadata=schema(title={group_name_title!r}))"


def get_model_group_info() -> t.Sequence[ModelGroupInfo]:
    all_group_models = (
        "photon_generation",
        "optics",
        "phasing",
        "charge_generation",
        "charge_collection",
        "charge_measurement",
        "readout_electronics",
        "charge_transfer",
        "signal_transfer",
    )

    lst = []  # type: t.List[ModelGroupInfo]

    for group_name in all_group_models:
        lst.append(
            ModelGroupInfo(
                name=group_name, class_name=under2camel(f"model_group_{group_name}")
            )
        )

    return lst


@functools.cache
def create_klass(cls: t.Union[t.Type, str]) -> Klass:
    import pyxel.detectors

    if isinstance(cls, str):
        cls_type = getattr(pyxel.detectors, cls)  # type: t.Type
        return create_klass(cls_type)

    # Try to find a base class
    _, *base_classes, _ = inspect.getmro(cls)  # type: tuple

    if base_classes:
        return Klass(cls, base_cls=base_classes[0])

    return Klass(cls)


def create_graph(cls: t.Type, graph: t.Mapping[Klass, t.Set[Klass]]) -> None:
    klass = create_klass(cls)  # type: Klass

    doc = get_doc_from_klass(klass)  # type: FuncDocumentation

    if klass.base_cls is None:
        parameters = (
            doc.parameters
        )  # type: t.Mapping[str, t.Union[Param, ParamDefault]]
    else:
        create_graph(cls=klass.base_cls, graph=graph)
        klass_base = create_klass(klass.base_cls)  # type: Klass
        graph[klass].add(klass_base)

        klass_base_doc = get_doc_from_klass(klass_base)  # type:  FuncDocumentation

        parameters = {**doc.parameters, **klass_base_doc.parameters}

    for name, parameter in parameters.items():
        assert parameter.annotation

        klass_param = create_klass(parameter.annotation)  # type: Klass
        graph[klass].add(klass_param)

        if klass_param.base_cls is not None:
            klass_param_base = create_klass(klass_param.base_cls)  # type: Klass
            graph[klass_param].add(klass_param_base)


def generate_detectors() -> t.Iterator[str]:
    from pyxel.detectors import APD, CCD, CMOS, MKID

    registered_detectors = (CCD, CMOS, MKID, APD)  # type: t.Sequence[t.Type[Detector]]

    # Build a dependency graph
    graph = defaultdict(set)  # type: t.Mapping[Klass, t.Set[Klass]]
    for detector in registered_detectors:  # type: t.Type[Detector]
        create_graph(cls=detector, graph=graph)

    # Generate code based on the dependency graph
    ts = TopologicalSorter(graph)
    for klass in ts.static_order():
        yield from generate_class(klass)

    yield ""
    yield "#"
    yield "# Outputs"
    yield "#"
    yield "ValidName = typing.Literal["
    yield "    'detector.image.array', 'detector.signal.array', 'detector.pixel.array'"
    yield "]"
    yield "ValidFormat = typing.Literal['fits', 'hdf', 'npy', 'txt', 'csv', 'png']"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class Outputs:"
    yield "    output_folder: pathlib.Path"
    yield "    save_data_to_file: typing.Optional["
    yield "        typing.Sequence[typing.Mapping[ValidName, typing.Sequence[ValidFormat]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Exposure"
    yield "#"
    yield "@dataclass"
    yield "class ExposureOutputs(Outputs):"
    yield "    save_exposure_data: typing.Optional["
    yield "        typing.Sequence[typing.Mapping[str, typing.Sequence[str]]]"
    yield "    ] = None"

    yield "@dataclass"
    yield "class Readout:"
    yield "    times: typing.Optional[typing.Union[typing.Sequence, str]] = None"
    yield "    times_from_file: typing.Optional[str] = None"
    yield "    start_time: float = 0.0"
    yield "    non_destructive: bool = False"
    yield ""
    yield ""
    yield "@schema(title='Exposure')"
    yield "@dataclass"
    yield "class Exposure:"
    yield "    outputs: ExposureOutputs"
    yield "    readout: Readout = field(default_factory=Readout)"
    yield "    result_type: typing.Literal['image', 'signal', 'pixel', 'all'] = 'all'"
    yield "    pipeline_seed: typing.Optional[int] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Observation"
    yield "#"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ObservationOutputs(Outputs):"
    yield "    save_observation_data: typing.Optional["
    yield "        typing.Sequence[typing.Mapping[str, typing.Sequence[str]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ParameterValues:"
    yield "    key: str"
    yield "    values: typing.Union["
    yield "        typing.Literal['_'],"
    yield "        typing.Sequence[typing.Literal['_']],"
    yield "        typing.Sequence[typing.Union[int, float]],"
    yield "        typing.Sequence[str],"
    yield "    ]"
    yield "    boundaries: typing.Optional[typing.Tuple[float, float]] = None"
    yield "    enabled: bool = True"
    yield "    logarithmic: bool = False"
    yield ""
    yield ""
    yield "@schema(title='Observation')"
    yield "@dataclass"
    yield "class Observation:"
    yield "    outputs: ObservationOutputs"
    yield "    parameters: typing.Sequence[ParameterValues]"
    yield "    readout: Readout"
    yield "    mode: str = 'product'"
    yield "    from_file: typing.Optional[str] = None"
    yield "    column_range: typing.Optional[typing.Tuple[int, int]] = None"
    yield "    with_dask: bool = False"
    yield "    result_type: typing.Literal['image', 'signal', 'pixel', 'all'] = 'all'"
    yield "    pipeline_seed: typing.Optional[int] = None"
    yield ""
    yield ""
    yield "#"
    yield "# Calibration"
    yield "#"
    yield "@dataclass"
    yield "class CalibrationOutputs(Outputs):"
    yield "    save_calibration_data: typing.Optional["
    yield "        typing.Sequence[typing.Mapping[str, typing.Sequence[str]]]"
    yield "    ] = None"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class Algorithm:"
    yield "    type: typing.Literal['sade', 'sga', 'nlopt'] = 'sade'"
    yield "    generations: int = 1"
    yield "    population_size: int = 1"

    yield "    # SADE #####"
    yield "    variant: int = 2"
    yield "    variant_adptv: int = 1"
    yield "    ftol: float = 1e-6"
    yield "    xtol: float = 1e-6"
    yield "    memory: bool = False"

    yield "    # SGA #####"
    yield "    cr: float = 0.9"
    yield "    eta_c: float = 1.0"
    yield "    m: float = 0.02"
    yield "    param_m: float = 1.0"
    yield "    param_s: int = 2"
    yield "    crossover: typing.Literal['single', 'exponential', 'binominal', 'sbx'] = 'exponential'"
    yield "    mutation: typing.Literal['uniform', 'gaussian', 'polynomial'] = 'polynomial'"
    yield "    selection: typing.Literal['tournament', 'truncated'] = 'tournament'"

    yield "    # NLOPT #####"
    yield "    nlopt_solver: typing.Literal["
    yield "        'cobyla', 'bobyqa', 'newuoa', 'newuoa_bound', 'praxis', 'neldermead',"
    yield "        'sbplx', 'mma', 'ccsaq', 'slsqp', 'lbfgs', 'tnewton_precond_restart',"
    yield "        'tnewton_precond', 'tnewton_restart', 'tnewton', 'var2', 'var1', 'auglag',"
    yield "        'auglag_eq'"
    yield "    ] = 'neldermead'"
    yield "    maxtime: int = 0"
    yield "    maxeval: int = 0"
    yield "    xtol_rel: float = 1.0e-8"
    yield "    xtol_abs: float = 0.0"
    yield "    ftol_rel: float = 0.0"
    yield "    ftol_abs: float = 0.0"
    yield "    stopval: typing.Optional[float] = None"
    yield "    # local_optimizer: typing.Optional['pg.nlopt'] = None"
    yield "    replacement: typing.Literal['best', 'worst', 'random'] = 'best'"
    yield "    nlopt_selection: typing.Literal['best', 'worst', 'random'] = 'best'"
    yield ""
    yield ""
    yield "@schema(title='Calibration')"
    yield "@dataclass"
    yield "class Calibration:"
    yield "    outputs: CalibrationOutputs"
    yield "    readout: Readout = field(default_factory=Readout)"
    yield "    output_dir: typing.Optional[pathlib.Path] = None"
    yield "    # fitting: typing.Optional[ModelFitting] = None"
    yield ""
    yield "    mode: typing.Literal['pipeline', 'single_model'] = 'pipeline'"
    yield "    result_type: typing.Literal['image', 'signal', 'pixel'] = 'image'"
    yield "    result_fit_range: typing.Optional[typing.Sequence[int]] = None"
    yield "    result_input_arguments: typing.Optional[typing.Sequence[ParameterValues]] = None"
    yield "    target_data_path: typing.Optional[typing.Sequence[pathlib.Path]] = None"
    yield "    target_fit_range: typing.Optional[typing.Sequence[int]] = None"
    yield "    fitness_function: typing.Optional[ModelFunction] = None"
    yield "    algorithm: typing.Optional[Algorithm] = None"
    yield "    parameters: typing.Optional[typing.Sequence[ParameterValues]] = None"
    yield "    pygmo_seed: typing.Optional[int] = None"
    yield "    pipeline_seed: typing.Optional[int] = None"
    yield "    num_islands: int = 1"
    yield "    num_evolutions: int = 1"
    yield "    num_best_decisions: typing.Optional[int] = None"
    yield "    topology: typing.Literal['unconnected', 'ring', 'fully_connected'] = 'unconnected'"
    yield "    type_islands: typing.Literal["
    yield "        'multiprocessing', 'multithreading', 'ipyparallel'"
    yield "    ] = 'multiprocessing'"
    yield "    weights_from_file: typing.Optional[typing.Sequence[pathlib.Path]] = None"
    yield "    weights: typing.Optional[typing.Sequence[float]] = None"
    yield ""
    yield ""

    # Create wrappers for the detectors
    detector_classes = ["CCD", "CMOS", "MKID", "APD"]  # type: t.Sequence[str]
    # for klass_name in detector_classes:
    #     yield f"#@schema(title='{klass_name}')"
    #     yield "#@dataclass"
    #     yield f"#class Wrapper{klass_name}:"
    #     yield f"#    {klass_name.lower()}: {klass_name}"
    #     yield ""
    #     yield ""

    # Create wrappers for the modes
    mode_classes = ["Exposure", "Observation", "Calibration"]  # type: t.Sequence[str]
    # for klass_name in mode_classes:
    #     yield f"#@schema(title='{klass_name}')"
    #     yield "#@dataclass"
    #     yield f"#class Wrapper{klass_name}:"
    #     yield f"#    {klass_name.lower()}: {klass_name}"
    #     yield ""
    #     yield ""

    # wrapper_detector_classes = [f"Wrapper{el}" for el in detector_classes]
    # wrapper_mode_classes = [f"Wrapper{el}" for el in mode_classes]

    yield "@dataclass"
    yield "class Configuration:"
    yield "    pipeline: DetailedDetectionPipeline"
    # yield f"    # mode: typing.Union[{', '.join(wrapper_mode_classes)}]"
    # yield f"    # detector: typing.Union[{', '.join(wrapper_detector_classes)}]"

    yield ""
    yield "    # Running modes"
    for klass_name in mode_classes:
        yield f"    {klass_name.lower()}: typing.Optional[{klass_name}] = field(default=None, metadata=schema(title={klass_name!r}))"

    yield ""
    yield "    # Detectors"
    for klass_name in detector_classes:
        yield f"    {klass_name.lower()}_detector: typing.Optional[{klass_name}] = field(default=None, metadata=schema(title={klass_name!r}))"


def generate_all_models() -> t.Iterator[str]:
    lst = get_model_group_info()
    yield "#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022."
    yield "#"
    yield "#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which"
    yield "#  is part of this Pyxel package. No part of the package, including"
    yield "#  this file, may be copied, modified, propagated, or distributed except according to"
    yield "#  the terms contained in the file ‘LICENCE.txt’."
    yield ""
    yield "######################################"
    yield "# Note: This code is auto-generated. #"
    yield "#       Don't modify it !            #"
    yield "######################################"
    yield ""

    yield "import pathlib"
    yield "import typing"
    yield "from dataclasses import dataclass, field"
    yield ""
    yield "import click"
    yield "from apischema import schema"
    yield ""
    yield ""

    yield "@dataclass"
    yield "class ModelFunction:"
    yield "    name: str"
    yield "    func: str = field(metadata=schema(pattern='^(?!pyxel\\.models\\.)'))"
    yield "    arguments: typing.Mapping[str, typing.Any] = field(default_factory=dict)"
    yield "    enabled: bool = True"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class ModelGroup:"
    yield "    models: typing.Sequence[ModelFunction]"
    yield "    name: str"
    yield ""
    yield ""
    yield "@dataclass"
    yield "class DetectionPipeline:"
    yield "    photon_generation: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    optics: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    phasing: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    charge_generation: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    charge_collection: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    charge_measurement: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    readout_electronics: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    charge_transfer: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield "    signal_transfer: typing.Sequence[ModelFunction] = field(default_factory=list)"
    yield ""
    yield ""

    yield from generate_group(lst)
    yield from generate_detectors()

    yield ""
    yield ""
    yield "@click.command()"
    yield "@click.option("
    yield "    '-f',"
    yield "    '--filename',"
    yield "    default='../../static/pyxel_schema.json',"
    yield "    type=click.Path(),"
    yield "    help='JSON schema filename',"
    yield "    show_default=True,"
    yield ")"
    yield "def create_json_schema(filename: pathlib.Path):"
    yield "    import json"
    yield ""
    yield "    from apischema.json_schema import JsonSchemaVersion, deserialization_schema"
    yield "    dct_schema = deserialization_schema("
    yield "        Configuration, version=JsonSchemaVersion.DRAFT_7"
    yield "    )"
    yield ""
    yield "    print(json.dumps(dct_schema))"
    yield ""
    yield "    full_filename = pathlib.Path(filename).resolve()"
    yield ""
    yield "    with full_filename.open('w') as fh:"
    yield "        json.dump(obj=dct_schema, fp=fh, indent=2)"
    yield ""
    yield ""
    yield "if __name__ == '__main__':"
    yield "    create_json_schema()"


def create_auto_generated(filename: Path) -> None:
    """Create an auto-generated file."""
    with Path(filename).open("w") as fh:
        for line in tqdm(generate_all_models()):
            fh.write(f"{line}\n")


@click.command()
@click.option(
    "-f",
    "--filename",
    default="./auto_generated.py",
    type=click.Path(),
    help="Auto generated filename.",
    show_default=True,
)
@click.version_option(version=__version__)
def main(filename: Path):
    """Create an auto-generated file."""
    create_auto_generated(filename)


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import typing as t

import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.optics import optical_psf
from pyxel.models.optics.poppy import (
    CircularAperture,
    HexagonAperture,
    MultiHexagonalAperture,
    RectangleAperture,
    SecondaryObscuration,
    SineWaveWFE,
    SquareAperture,
    ThinLens,
    ZernikeWFE,
    create_optical_parameter,
)


@pytest.fixture
def ccd_3x3() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=3,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )


@pytest.mark.parametrize(
    "dct, exp_parameter",
    [
        pytest.param(
            {}, None, marks=pytest.mark.xfail(raises=KeyError, strict=True), id="Empty"
        ),
        pytest.param(
            {"item": "CircularAperture", "radius": 3.14},
            CircularAperture(radius=3.14),
            id="CircularAperture",
        ),
        pytest.param(
            {"item": "ThinLens", "nwaves": 1.1, "radius": 2.2},
            ThinLens(nwaves=1.1, radius=2.2),
            id="ThinLens",
        ),
        pytest.param(
            {"item": "SquareAperture", "size": 1},
            SquareAperture(size=1),
            id="SquareAperture",
        ),
        pytest.param(
            {"item": "RectangularAperture", "width": 1.1, "height": 2.2},
            RectangleAperture(width=1.1, height=2.2),
            id="RectangularAperture",
        ),
        pytest.param(
            {"item": "HexagonAperture", "side": 1.1},
            HexagonAperture(side=1.1),
            id="HexagonAperture",
        ),
        pytest.param(
            {"item": "MultiHexagonalAperture", "side": 1.1, "rings": 2, "gap": 3.3},
            MultiHexagonalAperture(side=1.1, rings=2, gap=3.3),
            id="MultiHexagonalAperture",
        ),
        pytest.param(
            {
                "item": "SecondaryObscuration",
                "secondary_radius": 1.1,
                "n_supports": 2,
                "support_width": 3.3,
            },
            SecondaryObscuration(secondary_radius=1.1, n_supports=2, support_width=3.3),
            id="SecondaryObscuration",
        ),
        pytest.param(
            {
                "item": "ZernikeWFE",
                "radius": 1.1,
                "coefficients": [2.2, 3.3],
                "aperture_stop": 4.4,
            },
            ZernikeWFE(radius=1.1, coefficients=[2.2, 3.3], aperture_stop=4.4),
            id="ZernikeWFE",
        ),
        pytest.param(
            {
                "item": "SineWaveWFE",
                "spatialfreq": 2.2,
                "amplitude": 3.3,
                "rotation": 4.4,
            },
            SineWaveWFE(spatialfreq=2.2, amplitude=3.3, rotation=4.4),
            id="SineWaveWFE",
        ),
    ],
)
def test_create_optical_parameter(dct: t.Mapping, exp_parameter):
    """Test function 'create_optical_parameter'."""
    parameter = create_optical_parameter(dct)

    assert parameter == exp_parameter


@pytest.mark.parametrize(
    "wavelength, fov_arcsec, pixelscale, optical_system",
    [
        pytest.param(
            0.6e-6, 5, 0.01, [{"item": "CircularAperture", "radius": 3.0}], id="valid"
        ),
        pytest.param(
            -1,
            5,
            0.01,
            [{"item": "CircularAperture", "radius": 3.0}],
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="Negative 'wavelength'",
        ),
        pytest.param(
            0.6e-6,
            -1,
            0.01,
            [{"item": "CircularAperture", "radius": 3.0}],
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="Negative 'fov_arcsec'",
        ),
        pytest.param(
            0.6e-6,
            5,
            -1,
            [{"item": "CircularAperture", "radius": 3.0}],
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="Negative 'pixelscale'",
        ),
    ],
)
def test_optical_psf(
    ccd_3x3: CCD,
    wavelength: float,
    fov_arcsec: float,
    pixelscale: float,
    optical_system: t.Sequence[t.Mapping],
):
    """Test input parameters for function 'optical_psf'."""
    optical_psf(
        detector=ccd_3x3,
        wavelength=wavelength,
        fov_arcsec=fov_arcsec,
        pixelscale=pixelscale,
        optical_system=optical_system,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy

import pytest

from pyxel.detectors import Environment


@pytest.mark.parametrize(
    "temperature ",
    [0.0, 1000.0],
)
def test_create_environment(temperature: float):
    """Test when creating a valid `Environment` object."""
    _ = Environment(temperature=temperature)


@pytest.mark.parametrize(
    "temperature, exp_exc, exp_error",
    [
        pytest.param(-0.1, ValueError, r"\'temperature\'"),
        pytest.param(1001, ValueError, r"\'temperature\'"),
    ],
)
def test_create_invalid_environment(temperature: float, exp_exc, exp_error):
    """Test when creating an invalid `Environment` object."""
    with pytest.raises(exp_exc, match=exp_error):
        _ = Environment(temperature=temperature)


@pytest.mark.parametrize(
    "other_obj, is_equal",
    [
        pytest.param(None, False, id="None"),
        pytest.param(Environment(), False, id="Empty 'Environment'"),
        pytest.param(Environment(temperature=100.1), True, id="Valid"),
    ],
)
def test_is_equal(other_obj, is_equal):
    """Test equality statement for Environment."""
    obj = Environment(temperature=100.1)

    if is_equal:
        assert obj == other_obj
    else:
        assert obj != other_obj


@pytest.mark.parametrize(
    "obj, exp_dict",
    [
        (
            Environment(temperature=100.1),
            {"temperature": 100.1},
        )
    ],
)
def test_to_and_from_dict(obj, exp_dict):
    """Test methods 'to_dict' and 'from_dict'"""

    assert type(obj) == Environment

    # Convert from `Environment` to a `dict`
    dct = obj.to_dict()
    assert dct == exp_dict

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    assert copied_dct == exp_dict

    # Convert from `dict` to `Environment`
    other_obj = Environment.from_dict(copied_dct)
    assert type(other_obj) == Environment
    assert obj == other_obj
    assert obj is not other_obj
    assert copied_dct == exp_dict



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#

import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment


@pytest.fixture
def environment() -> Environment:
    return Environment()


@pytest.fixture
def ccd_geometry() -> CCDGeometry:
    return CCDGeometry(row=1, col=1)


@pytest.fixture
def ccd_characteristics() -> CCDCharacteristics:
    return CCDCharacteristics()


@pytest.fixture
def CCD_empty(ccd_geometry, environment, ccd_characteristics) -> CCD:
    detector = CCD(
        geometry=ccd_geometry,
        environment=environment,
        characteristics=ccd_characteristics,
    )
    return detector



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import typing as t

import numpy as np
import pytest
from typing_extensions import Literal

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDCharacteristics,
    CCDGeometry,
    CMOSCharacteristics,
    CMOSGeometry,
    Environment,
)
from pyxel.models.charge_transfer import cdm


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=273.15),
        characteristics=CCDCharacteristics(),
    )


@pytest.fixture
def input_array() -> np.ndarray:
    """Create a valid input_array."""
    out = np.ones((5, 5), dtype=float) * 1000
    out[:3, :] *= 0
    return out


@pytest.mark.parametrize(
    """
    direction,
    trap_release_times, 
    trap_densities, 
    sigma, 
    beta, 
    full_well_capacity, 
    max_electron_volume, 
    transfer_period, 
    charge_injection, 
    exp_err, 
    exp_exc
    """,
    [
        pytest.param(
            "parallel",
            [1.0],
            [1.0],
            [1.0],
            1.0,
            1.0,
            2.0,
            1.0,
            False,
            ValueError,
            r"'max_electron_volume' must be between 0.0 and 1.0.",
            id="Max volume out of bonds.",
        ),
        pytest.param(
            "parallel",
            [1.0],
            [1.0],
            [1.0],
            2.0,
            1.0,
            1.0,
            1.0,
            False,
            ValueError,
            r"'beta' must be between 0.0 and 1.0.",
            id="Beta out of bonds.",
        ),
        pytest.param(
            "parallel",
            [1.0],
            [1.0],
            [1.0],
            1.0,
            1.0e8,
            1.0,
            1.0,
            False,
            ValueError,
            "'full_well_capacity' must be between 0 and 1e7.",
            id="Fwc out of bonds.",
        ),
        pytest.param(
            "parallel",
            [1.0],
            [1.0],
            [1.0],
            1.0,
            1.0,
            1.0,
            20.0,
            False,
            ValueError,
            r"'transfer_period' must be between 0.0 and 10.0.",
            id="Transfer period out of bonds.",
        ),
        pytest.param(
            "parallel",
            [1.0, 2.0],
            [1.0],
            [1.0],
            1.0,
            1.0,
            1.0,
            1.0,
            False,
            ValueError,
            r"Length of 'sigma', 'trap_densities' and 'trap_release_times' not the same!",
            id="Different lengths.",
        ),
        pytest.param(
            "parallel",
            [],
            [],
            [],
            1.0,
            1.0,
            1.0,
            1.0,
            False,
            ValueError,
            r"Expecting inputs for at least one trap species.",
            id="Empty.",
        ),
    ],
)
def test_cdm_bad_inputs(
    ccd_5x5: CCD,
    direction: Literal["parallel", "serial"],
    beta: float,
    trap_release_times: t.Sequence[float],
    trap_densities: t.Sequence[float],
    sigma: t.Sequence[float],
    full_well_capacity: t.Optional[float],
    max_electron_volume: float,
    transfer_period: float,
    charge_injection: bool,
    exp_err,
    exp_exc,
):
    """Test function 'cdm' with bad inputs."""
    with pytest.raises(expected_exception=exp_err, match=exp_exc):
        cdm(
            detector=ccd_5x5,
            direction=direction,
            trap_release_times=trap_release_times,
            trap_densities=trap_densities,
            sigma=sigma,
            beta=beta,
            full_well_capacity=full_well_capacity,
            max_electron_volume=max_electron_volume,
            transfer_period=transfer_period,
            charge_injection=charge_injection,
        )


def test_cdm_with_cmos():
    """Test function 'cdm' with CMOS."""

    detector = CMOS(
        geometry=CMOSGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CMOSCharacteristics(),
    )

    with pytest.raises(
        expected_exception=TypeError, match="Expecting a `CCD` object for 'detector'."
    ):
        cdm(
            detector=detector,
            direction="parallel",
            trap_release_times=[1.0],
            trap_densities=[1.0],
            sigma=[1.0],
            beta=1.0,
            full_well_capacity=1.0,
            max_electron_volume=1.0,
            transfer_period=1.0,
            charge_injection=False,
        )


@pytest.mark.parametrize(
    """
    direction,
    trap_release_times, 
    trap_densities, 
    sigma, 
    beta, 
    full_well_capacity, 
    max_electron_volume, 
    transfer_period, 
    charge_injection, 
    """,
    [
        pytest.param(
            "parallel",
            [5.0e-3, 5.0e-3],
            [10.0, 10.0],
            [1.0e-15, 1e-15],
            0.3,
            10000.0,
            1.5e-10,
            1.0e-3,
            True,
        ),
    ],
)
def test_cdm_parallel(
    ccd_5x5: CCD,
    input_array: np.ndarray,
    direction: Literal["parallel", "serial"],
    beta: float,
    trap_release_times: t.Sequence[float],
    trap_densities: t.Sequence[float],
    sigma: t.Sequence[float],
    full_well_capacity: t.Optional[float],
    max_electron_volume: float,
    transfer_period: float,
    charge_injection: bool,
):
    """Test function 'cdm' with valid inputs."""
    expected = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [949.71893729, 949.71893729, 949.71893729, 949.71893729, 949.71893729],
            [983.92354165, 983.92354165, 983.92354165, 983.92354165, 983.92354165],
        ]
    )
    detector = ccd_5x5
    detector.pixel.array = input_array

    cdm(
        detector=detector,
        direction=direction,
        trap_release_times=trap_release_times,
        trap_densities=trap_densities,
        sigma=sigma,
        beta=beta,
        full_well_capacity=full_well_capacity,
        max_electron_volume=max_electron_volume,
        transfer_period=transfer_period,
        charge_injection=charge_injection,
    )

    np.testing.assert_array_almost_equal(detector.pixel.array, expected)


@pytest.mark.parametrize(
    """
    direction,
    trap_release_times, 
    trap_densities, 
    sigma, 
    beta, 
    full_well_capacity, 
    max_electron_volume, 
    transfer_period, 
    charge_injection, 
    """,
    [
        pytest.param(
            "serial",
            [5.0e-3, 5.0e-3],
            [10.0, 10.0],
            [1.0e-15, 1e-15],
            0.3,
            10000.0,
            1.5e-10,
            1.0e-3,
            None,
        ),
    ],
)
def test_cdm_serial(
    ccd_5x5: CCD,
    input_array: np.ndarray,
    direction: Literal["parallel", "serial"],
    beta: float,
    trap_release_times: t.Sequence[float],
    trap_densities: t.Sequence[float],
    sigma: t.Sequence[float],
    full_well_capacity: t.Optional[float],
    max_electron_volume: float,
    transfer_period: float,
    charge_injection: bool,
):
    """Test function 'cdm' with valid inputs."""
    expected = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [1000.0, 989.46583672, 986.54300532, 985.74893059, 985.56511735],
            [1000.0, 989.46583672, 986.54300532, 985.74893059, 985.56511735],
        ]
    )
    detector = ccd_5x5
    detector.pixel.array = input_array

    cdm(
        detector=detector,
        direction=direction,
        trap_release_times=trap_release_times,
        trap_densities=trap_densities,
        sigma=sigma,
        beta=beta,
        full_well_capacity=full_well_capacity,
        max_electron_volume=max_electron_volume,
        transfer_period=transfer_period,
        charge_injection=charge_injection,
    )

    np.testing.assert_array_almost_equal(detector.pixel.array, expected)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Badge creator script.
by David Lucsanyi
"""
import argparse
import subprocess


def badges():
    """Create badges."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-cov",
        "--coverage",
        type=float,
        default=None,
        help="Test coverage badge with percent",
    )
    parser.add_argument(
        "-ver", "--version", type=str, default=None, help="Version badge"
    )
    parser.add_argument(
        "-lic", "--license", type=str, default=None, help="License badge"
    )
    parser.add_argument(
        "-doc", "--documentation", type=str, default=None, help="Documentation badge"
    )
    parser.add_argument("-ch", "--chat", type=str, default=None, help="Chat badge")
    opts = parser.parse_args()

    if opts.coverage:
        if 0.0 <= opts.coverage < 20.0:
            colour = "red"
        elif 20.0 <= opts.coverage < 40.0:
            colour = "orange"
        elif 40.0 <= opts.coverage < 60.0:
            colour = "yellow"
        elif 60.0 <= opts.coverage < 80.0:
            colour = "yellowgreen"
        elif 80.0 <= opts.coverage < 90.0:
            colour = "green"
        elif 90.0 <= opts.coverage <= 100.0:
            colour = "brightgreen"
        else:
            colour = "lightgrey"
        percent = "{:.2f}".format(opts.coverage)
        process = (
            "wget -O coverage.svg https://img.shields.io/badge/coverage-"
            + percent
            + "%25-"
            + colour
            + ".svg"
        )
        subprocess.run(process, shell=True)

    if opts.documentation:
        colour = "brightgreen"
        process = (
            "wget -O documentation.svg https://img.shields.io/badge/docs-"
            + opts.documentation
            + "-"
            + colour
            + ".svg"
        )
        subprocess.run(process, shell=True)

    # if opts.license:
    #     colour = 'green'
    #     process = 'wget -O license.svg https://img.shields.io/badge/license-' + \
    #               opts.license + '-' + colour + '.svg'
    #     subprocess.run(process, shell=True)

    # if opts.chat:
    #     colour = 'green'
    #     process = 'wget -O chat.svg https://img.shields.io/badge/chat-' + \
    #               opts.chat + '-' + colour + '.svg'
    #     subprocess.run(process, shell=True)


if __name__ == "__main__":
    badges()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.readout_electronics import ac_crosstalk


@pytest.fixture
def ccd_8x8() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=8,
            col=8,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )


@pytest.mark.parametrize(
    "coupling_matrix, channel_matrix, readout_directions",
    [
        pytest.param(
            [[1, 0.5, 0, 0], [0.5, 1, 0, 0], [0, 0, 1, 0.5], [0, 0, 0.5, 1]],
            [1, 2, 3, 4],
            [1, 2, 1, 2],
            id="4 channels",
        ),
        pytest.param([[1, 0.5], [0.5, 1]], [1, 2], [1, 2], id="2 channels"),
        pytest.param([[1]], [1], [1], id="1 channel"),
    ],
)
def test_ac_crosstalk(
    ccd_8x8: CCD, coupling_matrix, channel_matrix, readout_directions
):
    """Test model 'ac_crosstalk' with valid parameters."""
    ac_crosstalk(
        detector=ccd_8x8,
        coupling_matrix=coupling_matrix,
        channel_matrix=channel_matrix,
        readout_directions=readout_directions,
    )


# TODO: Add more tests with a 2D 'channel_matrix'
# TODO: Add more tests with file(s)
@pytest.mark.parametrize(
    "coupling_matrix, channel_matrix, readout_directions, exp_exc, exp_msg",
    [
        pytest.param(
            [1, 0.5, 0, 0, 0.5, 1, 0, 0, 0, 0, 1, 0.5, 0, 0, 0.5, 1],
            [1, 2, 3, 4],
            [1, 2, 1, 2],
            ValueError,
            "Expecting 2D 'coupling_matrix'",
            id="1D coupling_matrix",
        ),
        pytest.param(
            [[1, 0.5], [0.5, 1], [0, 0], [0, 0]],
            [1, 2, 3, 4],
            [1, 2, 1, 2],
            ValueError,
            "Expecting a matrix of 4x4 elements for 'coupling_matrix'",
            id="2x4 coupling_matrix",
        ),
        pytest.param(
            [[1, 0.5, 0], [0.5, 1, 0], [0, 0, 1]],
            [1, 2, 3],
            [1, 2, 1],
            ValueError,
            "Can't split detector array horizontally for a given number of amplifiers",
            id="3 channels",
        ),
        pytest.param(
            [[1, 0.5, 0, 0], [0.5, 1, 0, 0], [0, 0, 1, 0.5], [0, 0, 0.5, 1]],
            [1, 2, 3, 4],
            [1, 2, 1],
            ValueError,
            "Channel matrix and readout directions arrays not the same size",
            id="Different size",
        ),
    ],
)
def test_ac_crosstalk_invalid_params(
    ccd_8x8: CCD, coupling_matrix, channel_matrix, readout_directions, exp_exc, exp_msg
):
    """Test model 'ac_crosstalk' with invalid parameters."""
    with pytest.raises(exp_exc, match=exp_msg):
        ac_crosstalk(
            detector=ccd_8x8,
            coupling_matrix=coupling_matrix,
            channel_matrix=channel_matrix,
            readout_directions=readout_directions,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


import numpy as np
import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.readout_electronics import sar_adc_with_noise


@pytest.fixture
def ccd_10x3() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(
            adc_bit_resolution=8, adc_voltage_range=(0.0, 10.0)
        ),
    )


@pytest.mark.parametrize(
    "strengths, noises",
    [
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0], id="list only 0"
        ),
        pytest.param(
            np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=float),
            np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=float),
            id="array only 0.0",
        ),
        pytest.param(
            [10, 20, 30, 40, 50, 60, 70, 80], [1, 2, 3, 4, 5, 6, 7, 8], id="list"
        ),
    ],
)
def test_sar_adc_with_noise(ccd_10x3: CCD, strengths, noises):
    """Test model 'sar_adc_with_noise' with valid inputs."""
    sar_adc_with_noise(detector=ccd_10x3, strengths=strengths, noises=noises)


@pytest.mark.parametrize(
    "strengths, noises, exp_error, exp_msg",
    [
        pytest.param(
            [],
            [0, 0, 0, 0, 0, 0, 0, 0],
            ValueError,
            "parameter 'strengths'",
            id="No 'strengths",
        ),
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0, 0],
            [],
            ValueError,
            "parameter 'noises'",
            id="No 'noises",
        ),
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            ValueError,
            "parameter 'strengths'",
            id="Missing one 'strength",
        ),
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            ValueError,
            "parameter 'noises'",
            id="Missing one 'noise",
        ),
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            ValueError,
            "parameter 'strengths'",
            id="Extra 'strength",
        ),
        pytest.param(
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            ValueError,
            "parameter 'noises'",
            id="Extra 'noise",
        ),
    ],
)
def test_sar_adc_with_noise_invalid_params(
    ccd_10x3: CCD, strengths, noises, exp_error, exp_msg
):
    """Test model 'sar_adc_with_noise' with invalid inputs."""
    with pytest.raises(exp_error, match=exp_msg):
        sar_adc_with_noise(detector=ccd_10x3, strengths=strengths, noises=noises)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import os
import re
import typing as t
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from astropy.io import fits
from PIL import Image
from pytest_httpserver import HTTPServer  # pip install pytest-httpserver

import pyxel


@pytest.fixture
def valid_multiple_hdus() -> fits.HDUList:
    """Create a valid HDUList with one 'PrimaryHDU' and one 'ImageHDU'."""
    # Create a new image
    primary_2d = np.array([[5, 6], [7, 8]], dtype=np.uint16)
    secondary_2d = np.array([[9, 10], [11, 12]], dtype=np.uint16)

    hdu_primary = fits.PrimaryHDU(primary_2d)
    hdu_secondary = fits.ImageHDU(secondary_2d)
    hdu_lst = fits.HDUList([hdu_primary, hdu_secondary])

    return hdu_lst


@pytest.fixture
def valid_pil_image() -> Image.Image:
    """Create a valid RGB PIL image."""
    data_2d = np.array([[10, 20], [30, 40]], dtype="uint8")
    pil_image = Image.fromarray(data_2d).convert("RGB")

    return pil_image


@pytest.fixture
def valid_data2d_http_hostname(
    tmp_path: Path,
    httpserver: HTTPServer,
    valid_pil_image: Image.Image,
    valid_multiple_hdus: fits.HDUList,
) -> str:
    """Create valid 2D files on a temporary folder and HTTP server."""
    # Get current folder
    current_folder = Path().cwd()  # type: Path

    try:
        os.chdir(tmp_path)

        # Create folder 'data'
        Path("data").mkdir(parents=True, exist_ok=True)

        data_2d = np.array([[1, 2], [3, 4]], dtype=np.uint16)

        # Save 2d images
        fits.writeto("data/img.fits", data=data_2d)
        fits.writeto("data/img2.FITS", data=data_2d)
        valid_multiple_hdus.writeto("data/img_multiple.fits")
        valid_multiple_hdus.writeto("data/img_multiple2.FITS")
        np.save("data/img.npy", arr=data_2d)
        np.savetxt("data/img_tab.txt", X=data_2d, delimiter="\t")
        np.savetxt("data/img_space.txt", X=data_2d, delimiter=" ")
        np.savetxt("data/img_comma.txt", X=data_2d, delimiter=",")
        np.savetxt("data/img_pipe.txt", X=data_2d, delimiter="|")
        np.savetxt("data/img_semicolon.txt", X=data_2d, delimiter=";")

        valid_pil_image.save("data/img.jpg")
        valid_pil_image.save("data/img.jpeg")
        valid_pil_image.save("data/img.png")
        valid_pil_image.save("data/img2.PNG")
        valid_pil_image.save("data/img.tiff")
        valid_pil_image.save("data/img.tif")
        valid_pil_image.save("data/img.bmp")

        text_filenames = [
            "data/img_tab.txt",
            "data/img_space.txt",
            "data/img_comma.txt",
            "data/img_pipe.txt",
            "data/img_semicolon.txt",
        ]

        # Put text data in a fake HTTP server
        for filename in text_filenames:
            with open(filename, "r") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type="text/plain"
                )

        binary_filenames = [
            ("data/img.fits", "text/plain"),  # TODO: Change this type
            ("data/img2.FITS", "text/plain"),  # TODO: Change this type
            ("data/img_multiple.fits", "text/plain"),  # TODO: Change this type
            ("data/img_multiple2.FITS", "text/plain"),  # TODO: Change this type
            ("data/img.npy", "text/plain"),  # TODO: Change this type
            ("data/img.jpg", "image/jpeg"),
            ("data/img.jpeg", "image/jpeg"),
            ("data/img.png", "image/png"),
            ("data/img2.PNG", "image/png"),
            ("data/img.tif", "image/tiff"),
            ("data/img.tiff", "image/tiff"),
            ("data/img.bmp", "image/bmp"),
        ]

        # Put binary data in a fake HTTP server
        for filename, content_type in binary_filenames:
            with open(filename, "rb") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type=content_type
                )

        # Extract an url (e.g. 'http://localhost:59226/)
        url = httpserver.url_for("")  # type: str

        # Extract the hostname (e.g. 'localhost:59226')
        hostname = re.findall("http://(.*)/", url)[0]  # type: str

        yield hostname

    finally:
        os.chdir(current_folder)


@pytest.fixture
def invalid_data2d_hostname(tmp_path: Path, httpserver: HTTPServer) -> str:
    """Create invalid 2D files on a temporary folder and HTTP server."""
    # Get current folder
    current_folder = Path().cwd()  # type: Path

    try:
        os.chdir(tmp_path)

        # Create folder 'data'
        Path("invalid_data").mkdir(parents=True, exist_ok=True)

        data_2d = np.array([[1, 2], [3, 4]], dtype=np.uint16)

        np.savetxt("invalid_data/img_X.txt", X=data_2d, delimiter="X")

        text_filenames = ["invalid_data/img_X.txt"]

        # Put text data in a fake HTTP server
        for filename in text_filenames:
            with open(filename, "r") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type="text/plain"
                )

        # Extract an url (e.g. 'http://localhost:59226/)
        url = httpserver.url_for("")  # type: str

        # Extract the hostname (e.g. 'localhost:59226')
        hostname = re.findall("http://(.*)/", url)[0]  # type: str

        yield hostname

    finally:
        os.chdir(current_folder)


@pytest.fixture
def valid_table_http_hostname(tmp_path: Path, httpserver: HTTPServer) -> str:
    """Create valid tables locally and on a temporary HTTP server."""
    # Get current folder
    current_folder = Path().cwd()  # type: Path

    try:
        os.chdir(tmp_path)

        # Create folder 'data'
        Path("data").mkdir(parents=True, exist_ok=True)

        df = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype="float64")

        # Save tables
        df.to_csv("data/table_tab.txt", header=False, index=False, sep="\t")
        df.to_csv("data/table_space.TXT", header=False, index=False, sep=" ")
        df.to_csv("data/table_comma.data", header=False, index=False, sep=",")
        df.to_csv("data/table_pipe.txt", header=False, index=False, sep="|")
        df.to_csv("data/table_semicolon.txt", header=False, index=False, sep=";")

        df.to_excel("data/table.xlsx", header=False, index=False)
        np.save("data/table.npy", arr=df.to_numpy())

        text_filenames = [
            "data/table_tab.txt",
            "data/table_space.TXT",
            "data/table_comma.data",
            "data/table_pipe.txt",
            "data/table_semicolon.txt",
        ]

        # Put text data in a fake HTTP server
        for filename in text_filenames:
            with open(filename, "r") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type="text/plain"
                )

        binary_filenames = [
            ("data/table.xlsx", "application/octet-stream"),
            ("data/table.npy", "application/octet-stream"),
        ]

        # Put binary data in a fake HTTP server
        for filename, content_type in binary_filenames:
            with open(filename, "rb") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type=content_type
                )

        # Extract an url (e.g. 'http://localhost:59226/)
        url = httpserver.url_for("")  # type: str

        # Extract the hostname (e.g. 'localhost:59226')
        hostname = re.findall("http://(.*)/", url)[0]  # type: str

        yield hostname

    finally:
        os.chdir(current_folder)


@pytest.fixture
def invalid_table_http_hostname(tmp_path: Path, httpserver: HTTPServer) -> str:
    """Create invalid tables on temporary folder and HTTP server."""
    # Get current folder
    current_folder = Path().cwd()  # type: Path

    try:
        os.chdir(tmp_path)

        # Create folder 'data'
        Path("invalid_data").mkdir(parents=True, exist_ok=True)

        df = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype="float64")

        # Save tables
        df.to_csv("invalid_data/table_X.txt", header=False, index=False, sep="X")
        Path("invalid_data/table_empty.txt").touch()

        text_filenames = [
            "invalid_data/table_X.txt",
            "invalid_data/table_empty.txt",
        ]

        # Put text data in a fake HTTP server
        for filename in text_filenames:
            with open(filename, "r") as fh:
                response_data = fh.read()  # type: str
                httpserver.expect_request(f"/{filename}").respond_with_data(
                    response_data, content_type="text/plain"
                )

        # Extract an url (e.g. 'http://localhost:59226/)
        url = httpserver.url_for("")  # type: str

        # Extract the hostname (e.g. 'localhost:59226')
        hostname = re.findall("http://(.*)/", url)[0]  # type: str

        yield hostname

    finally:
        os.chdir(current_folder)


@pytest.mark.parametrize(
    "filename, exp_error, exp_message",
    [
        ("dummy", ValueError, "Image format not supported"),
        ("dummy.foo", ValueError, "Image format not supported"),
        ("unknown.fits", FileNotFoundError, None),
        (Path("unknown.fits"), FileNotFoundError, r"can not be found\.$"),
        ("https://domain/unknown.fits", FileNotFoundError, None),
        ("invalid_data/img_X.txt", ValueError, "Cannot find the separator"),
        (
            "http://{host}/invalid_data/img_X.txt",
            ValueError,
            "Cannot find the separator",
        ),
    ],
)
def test_invalid_filename(
    invalid_data2d_hostname: str,
    filename,
    exp_error: TypeError,
    exp_message: t.Optional[str],
):
    """Test invalid filenames."""
    if isinstance(filename, str):
        filename = filename.format(host=invalid_data2d_hostname)  # type: str

    with pytest.raises(exp_error, match=exp_message):
        _ = pyxel.load_image(filename)


@pytest.mark.parametrize("with_caching", [False, True])
@pytest.mark.parametrize(
    "filename, exp_data",
    [
        # FITS files
        ("data/img.fits", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img2.FITS", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img_multiple.fits", np.array([[5, 6], [7, 8]], np.uint16)),
        ("data/img_multiple2.FITS", np.array([[5, 6], [7, 8]], np.uint16)),
        (Path("data/img.fits"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("./data/img.fits"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_multiple.fits"), np.array([[5, 6], [7, 8]], np.uint16)),
        ("http://{host}/data/img.fits", np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_multiple.fits", np.array([[5, 6], [7, 8]], np.uint16)),
        (
            "http://{host}/data/img_multiple2.FITS",
            np.array([[5, 6], [7, 8]], np.uint16),
        ),
        # Numpy binary files
        ("data/img.npy", np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img.npy"), np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img.npy", np.array([[1, 2], [3, 4]], np.uint16)),
        # Numpy text files
        ("data/img_tab.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img_space.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img_comma.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img_pipe.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("data/img_semicolon.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_tab.txt"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_space.txt"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_comma.txt"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_pipe.txt"), np.array([[1, 2], [3, 4]], np.uint16)),
        (Path("data/img_semicolon.txt"), np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_tab.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_space.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_comma.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_pipe.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        ("http://{host}/data/img_semicolon.txt", np.array([[1, 2], [3, 4]], np.uint16)),
        # JPG files
        ("data/img.jpg", np.array([[13, 19], [28, 34]])),
        ("data/img.jpeg", np.array([[13, 19], [28, 34]])),
        ("http://{host}/data/img.jpg", np.array([[13, 19], [28, 34]])),
        ("http://{host}/data/img.jpeg", np.array([[13, 19], [28, 34]])),
        # PNG files
        ("data/img.png", np.array([[10, 20], [30, 40]])),
        ("data/img2.PNG", np.array([[10, 20], [30, 40]])),
        ("http://{host}/data/img.png", np.array([[10, 20], [30, 40]])),
        ("http://{host}/data/img2.PNG", np.array([[10, 20], [30, 40]])),
        # TIFF files
        ("data/img.tif", np.array([[10, 20], [30, 40]])),
        ("data/img.tiff", np.array([[10, 20], [30, 40]])),
        ("http://{host}/data/img.tif", np.array([[10, 20], [30, 40]])),
        ("http://{host}/data/img.tiff", np.array([[10, 20], [30, 40]])),
        # BMP files
        ("data/img.bmp", np.array([[10, 20], [30, 40]])),
    ],
)
def test_load_image(
    with_caching: bool,
    valid_data2d_http_hostname: str,
    filename: t.Union[str, Path],
    exp_data: np.ndarray,
):
    """Test function 'load_image' with local and remote files."""
    with pyxel.set_options(cache_enabled=with_caching):
        if isinstance(filename, Path):
            # Load data
            data_2d = pyxel.load_image(filename)
        else:
            full_url = filename.format(host=valid_data2d_http_hostname)  # type: str

            # Load data
            data_2d = pyxel.load_image(full_url)

        # Check 'data_2d
        np.testing.assert_equal(data_2d, exp_data)


@pytest.mark.parametrize(
    "filename, exp_error, exp_message",
    [
        ("dummy", ValueError, r"^Only (.*) implemented.$"),
        (
            Path("unknown.txt"),
            FileNotFoundError,
            f"^Input file (.*) can not be found\.$",
        ),
        ("unknown.txt", FileNotFoundError, None),
        ("invalid_data/table_X.txt", ValueError, "Cannot find the separator"),
        (
            "http://{host}/invalid_data/table_X.txt",
            ValueError,
            "Cannot find the separator",
        ),
        ("invalid_data/table_empty.txt", ValueError, "Cannot find the separator"),
        (
            "http://{host}/invalid_data/table_empty.txt",
            ValueError,
            "Cannot find the separator",
        ),
    ],
)
def test_load_table_invalid_filename(
    invalid_table_http_hostname: str,
    filename,
    exp_error: TypeError,
    exp_message: t.Optional[str],
):
    """Test function 'load_table' with invalid filenames."""
    if isinstance(filename, str):
        filename = filename.format(host=invalid_table_http_hostname)

    with pytest.raises(exp_error, match=exp_message):
        _ = pyxel.load_table(filename)


@pytest.mark.parametrize("with_caching", [False, True])
@pytest.mark.parametrize(
    "filename",
    [
        "data/table_tab.txt",
        "data/table_space.TXT",
        "data/table_comma.data",
        "data/table_pipe.txt",
        "data/table_semicolon.txt",
        "data/table.xlsx",
        "data/table.npy",
        Path("data/table_tab.txt"),
        Path("data/table_space.TXT"),
        Path("data/table_comma.data"),
        Path("data/table_pipe.txt"),
        Path("data/table_semicolon.txt"),
        Path("data/table.xlsx"),
        Path("data/table.npy"),
        "http://{host}/data/table_tab.txt",
        "http://{host}/data/table_space.TXT",
        "http://{host}/data/table_comma.data",
        "http://{host}/data/table_pipe.txt",
        "http://{host}/data/table_semicolon.txt",
        "http://{host}/data/table.xlsx",
        "http://{host}/data/table.npy",
    ],
)
def test_load_table(with_caching: bool, valid_table_http_hostname: str, filename):
    """Test function 'load_table'."""
    with pyxel.set_options(cache_enabled=with_caching):
        if isinstance(filename, Path):
            # Load data
            table = pyxel.load_table(filename)
        else:
            full_url = filename.format(host=valid_table_http_hostname)  # type: str

            # Load data
            table = pyxel.load_table(full_url)

        exp_table = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype="float64")
        pd.testing.assert_frame_equal(table, exp_table)


@pytest.mark.parametrize("filename", ["dummy.foo"])
def test_load_table_invalid_format(tmp_path: Path, filename: str):
    # Create an empty file
    full_filename = tmp_path.joinpath(filename)  # type: Path
    full_filename.touch()

    with pytest.raises(ValueError):
        _ = pyxel.load_table(full_filename)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import typing as t

import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.readout_electronics import simple_adc


@pytest.fixture
def ccd_3x3() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=3,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(
            adc_bit_resolution=16, adc_voltage_range=(0.0, 10.0)
        ),
    )
    return detector


@pytest.mark.parametrize("data_type", ["uint16", "uint32", "uint64", "uint"])
def test_simple_adc(ccd_3x3: CCD, data_type: str):
    """Test model 'simple_adc'."""
    simple_adc(detector=ccd_3x3, data_type=data_type)


@pytest.mark.parametrize(
    "data_type, exp_exc, exp_error",
    [
        ("float", TypeError, "Expecting a signed/unsigned integer"),
        ("float32", TypeError, "Expecting a signed/unsigned integer"),
        ("float64", TypeError, "Expecting a signed/unsigned integer"),
        ("uint8", TypeError, "Expected type of Image array is uint64"),
        ("int32", TypeError, "Expected type of Image array is uint64"),
        ("int64", TypeError, "Expected type of Image array is uint64"),
        ("int", TypeError, "Expected type of Image array is uint64"),
    ],
)
def test_simple_adc_wrong_data_type(
    ccd_3x3: CCD,
    data_type: str,
    exp_exc,
    exp_error,
):
    """Test model 'simple_adc' with wrong 'data_type'."""
    with pytest.raises(exp_exc, match=exp_error):
        simple_adc(
            detector=ccd_3x3,
            data_type=data_type,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import typing as t

import numpy as np
import pytest
from typing_extensions import Literal

from pyxel.util import fit_into_array


@pytest.fixture
def array_2d() -> np.ndarray:
    """Create 2D data."""
    array = np.array([[1, 2], [3, 4], [5, 6]])

    return array


@pytest.fixture
def array_1d_col() -> np.ndarray:
    """Create 2D data."""
    array = np.array([[1], [2], [3], [4], [5]])

    return array


@pytest.fixture
def array_1d_row() -> np.ndarray:
    """Create 2D data."""
    array = np.array([[1, 2, 3, 4, 5]])

    return array


@pytest.mark.parametrize(
    "output_shape, relative_position, align, allow_smaller_array, expected_output",
    [
        pytest.param(
            (5, 5),
            (0, 0),
            None,
            True,
            np.array(
                [
                    [1.0, 2.0, 0.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0, 0.0],
                    [5.0, 6.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (-1, -1),
            None,
            True,
            np.array(
                [
                    [4.0, 0.0, 0.0, 0.0, 0.0],
                    [6.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (3, 3),
            None,
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 2.0],
                    [0.0, 0.0, 0.0, 3.0, 4.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (0, 0),
            "center",
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 2.0, 0.0, 0.0],
                    [0.0, 3.0, 4.0, 0.0, 0.0],
                    [0.0, 5.0, 6.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (0, 0),
            "top_left",
            True,
            np.array(
                [
                    [1.0, 2.0, 0.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0, 0.0],
                    [5.0, 6.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (0, 0),
            "bottom_left",
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 2.0, 0.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0, 0.0],
                    [5.0, 6.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (0, 0),
            "bottom_right",
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 2.0],
                    [0.0, 0.0, 0.0, 3.0, 4.0],
                    [0.0, 0.0, 0.0, 5.0, 6.0],
                ]
            ),
        ),
        pytest.param(
            (5, 5),
            (0, 0),
            "top_right",
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0, 1.0, 2.0],
                    [0.0, 0.0, 0.0, 3.0, 4.0],
                    [0.0, 0.0, 0.0, 5.0, 6.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
    ],
)
def test_fit_into_array_2d(
    array_2d,
    output_shape: t.Tuple[int, int],
    relative_position: t.Tuple[int, int],
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    allow_smaller_array: bool,
    expected_output: np.ndarray,
):
    """Test function 'fit_into_array' with some valid inputs."""

    output = fit_into_array(
        array=array_2d,
        output_shape=output_shape,
        relative_position=relative_position,
        align=align,
        allow_smaller_array=allow_smaller_array,
    )

    np.testing.assert_array_almost_equal(output, expected_output)


@pytest.mark.parametrize(
    "output_shape, relative_position, align, allow_smaller_array, expected_output",
    [
        pytest.param(
            (5, 1),
            (0, 0),
            None,
            True,
            np.array([[1.0], [2.0], [3.0], [4.0], [5.0]]),
        ),
        pytest.param(
            (5, 3),
            (0, 0),
            None,
            True,
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [2.0, 0.0, 0.0],
                    [3.0, 0.0, 0.0],
                    [4.0, 0.0, 0.0],
                    [5.0, 0.0, 0.0],
                ]
            ),
        ),
        pytest.param(
            (5, 3),
            (0, 1),
            None,
            True,
            np.array(
                [
                    [0.0, 1.0, 0.0],
                    [0.0, 2.0, 0.0],
                    [0.0, 3.0, 0.0],
                    [0.0, 4.0, 0.0],
                    [0.0, 5.0, 0.0],
                ]
            ),
        ),
    ],
)
def test_fit_into_array_1d_col(
    array_1d_col,
    output_shape: t.Tuple[int, int],
    relative_position: t.Tuple[int, int],
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    allow_smaller_array: bool,
    expected_output: np.ndarray,
):
    """Test function 'fit_into_array' with some valid inputs."""

    output = fit_into_array(
        array=array_1d_col,
        output_shape=output_shape,
        relative_position=relative_position,
        align=align,
        allow_smaller_array=allow_smaller_array,
    )

    np.testing.assert_array_almost_equal(output, expected_output)


@pytest.mark.parametrize(
    "output_shape, relative_position, align, allow_smaller_array, expected_output",
    [
        pytest.param(
            (1, 5),
            (0, 0),
            None,
            True,
            np.array([[1.0, 2.0, 3.0, 4.0, 5.0]]),
        ),
        pytest.param(
            (5, 3),
            (4, 0),
            None,
            True,
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [1.0, 2.0, 3.0],
                ]
            ),
        ),
        pytest.param(
            (3, 5),
            (0, 0),
            None,
            True,
            np.array(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        ),
    ],
)
def test_fit_into_array_1d_row(
    array_1d_row,
    output_shape: t.Tuple[int, int],
    relative_position: t.Tuple[int, int],
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    allow_smaller_array: bool,
    expected_output: np.ndarray,
):
    """Test function 'fit_into_array' with some valid inputs."""

    output = fit_into_array(
        array=array_1d_row,
        output_shape=output_shape,
        relative_position=relative_position,
        align=align,
        allow_smaller_array=allow_smaller_array,
    )

    np.testing.assert_array_almost_equal(output, expected_output)


@pytest.mark.parametrize(
    "output_shape, relative_position, align, allow_smaller_array, exp_exc, exp_error",
    [
        pytest.param(
            (10, 10),
            (0, 0),
            "foo",
            True,
            ValueError,
            r"'.*' is not a valid Alignment",
        ),
        pytest.param(
            (20, 20),
            (0, 0),
            None,
            False,
            ValueError,
            "Input array too small to fit into the desired shape!.",
        ),
        pytest.param(
            (10, 10),
            (15, 15),
            None,
            True,
            ValueError,
            "No overlap of array and target in Y and X dimension.",
        ),
        pytest.param(
            (10, 10),
            (15, 0),
            None,
            True,
            ValueError,
            "No overlap of array and target in Y dimension.",
        ),
        pytest.param(
            (10, 10),
            (0, 15),
            None,
            True,
            ValueError,
            "No overlap of array and target in X dimension.",
        ),
    ],
)
def test_fit_into_array_bad_inputs(
    output_shape: t.Tuple[int, int],
    relative_position: t.Tuple[int, int],
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ],
    allow_smaller_array: bool,
    exp_exc,
    exp_error,
):
    """Test function 'fit_into_array' with bad inputs."""

    array = np.ones((10, 10))

    with pytest.raises(exp_exc, match=exp_error):
        fit_into_array(
            array=array,
            output_shape=output_shape,
            relative_position=relative_position,
            align=align,
            allow_smaller_array=allow_smaller_array,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import typing as t
from pathlib import Path

import pytest

import pyxel
from pyxel.options import GlobalOptions, global_options


def test_with_context_manager():
    """Test with embedded context managers."""
    assert global_options == GlobalOptions()

    with pyxel.set_options(cache_folder="dummy"):
        assert global_options == GlobalOptions(cache_folder="dummy")

        with pyxel.set_options(cache_folder=Path("foo"), cache_enabled=True):
            assert global_options == GlobalOptions(
                cache_folder=Path("foo"), cache_enabled=True
            )

        assert global_options == GlobalOptions(cache_folder="dummy")

    assert global_options == GlobalOptions()


def test_global():
    assert global_options == GlobalOptions()

    pyxel.set_options(cache_folder="dummy")
    assert global_options == GlobalOptions(cache_folder="dummy")

    pyxel.set_options(cache_folder=None)
    assert global_options == GlobalOptions()


@pytest.mark.parametrize(
    "dct, exp_error, exp_msg",
    [
        pytest.param(
            {"non_sense": "foo"}, KeyError, "Wrong option", id="Unknown option"
        ),
        pytest.param({"cache_enabled": "Hello"}, TypeError, None, id="Wrong type"),
        pytest.param({"cache_folder": 42}, TypeError, None, id="Wrong type"),
    ],
)
def test_with_wrong_inputs(
    dct: dict, exp_error: t.Type[TypeError], exp_msg: t.Optional[str]
):
    """Test with a bad input."""
    with pytest.raises(exp_error, match=exp_msg):
        pyxel.set_options(**dct)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.calibration.util import slice_to_range


@pytest.mark.parametrize(
    "data, exp_result",
    [
        pytest.param(slice(10), range(10), id="0 to 10"),
        pytest.param(slice(20, 30), range(20, 30), id="20 to 30"),
    ],
)
def test_slice_to_range(data, exp_result):
    """Test function 'slice_to_range' with valid inputs."""
    result = slice_to_range(data)
    assert result == exp_result


@pytest.mark.parametrize(
    "data, exp_exp, exp_msg",
    [
        pytest.param(
            slice(10, 20, 2),
            ValueError,
            "Cannot use parameter 'step' in the slice object",
            id="with_step",
        ),
        pytest.param(
            slice(5, None),
            ValueError,
            "Missing 'stop' parameter in the slice object",
            id="No 'stop'",
        ),
        pytest.param(
            slice(0),
            ValueError,
            "Parameter 'stop' must be strictly greater than 'start'",
            id="'stop' == 0",
        ),
        pytest.param(
            slice(5, 5),
            ValueError,
            "Parameter 'stop' must be strictly greater than 'start'",
            id="'stop' == 'start'",
        ),
        pytest.param(
            slice(5, 4),
            ValueError,
            "Parameter 'stop' must be strictly greater than 'start'",
            id="'stop' smaller than 'start'",
        ),
        pytest.param(
            slice(-1, 4),
            ValueError,
            "Parameter 'start' must be strictly positive",
            id="negative 'start'",
        ),
    ],
)
def test_slice_to_range_bad_inputs(data, exp_exp, exp_msg):
    """Test function 'slice_to_range' with invalid inputs."""
    with pytest.raises(exp_exp, match=exp_msg):
        _ = slice_to_range(data)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Model for loading PSF from file."""
import typing as t
from pathlib import Path

import numpy as np
from astropy.convolution import convolve_fft

from pyxel.detectors import Detector
from pyxel.inputs import load_image


def apply_psf(
    array: np.ndarray, psf: np.ndarray, normalize_kernel: bool = True
) -> np.ndarray:
    """Convolve the input array with the point spread function kernel.

    Parameters
    ----------
    array: ndarray
        Input array.
    psf: ndarray
        Convolution kernel.
    normalize_kernel: bool
        Normalize kernel.

    Returns
    -------
    array_2d: ndarray
    """

    mean = np.mean(array)

    array_2d = convolve_fft(
        array,
        kernel=psf,
        boundary="fill",
        fill_value=mean,
        normalize_kernel=normalize_kernel,
    )

    return array_2d


def load_psf(
    detector: Detector, filename: t.Union[str, Path], normalize_kernel: bool = True
) -> None:
    """Load a point spread function from file and convolve the photon array with the PSF.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    filename: Path or str
        Input filename of the point spread function.
    normalize_kernel: bool
        Normalize kernel.
    """
    psf = load_image(filename)

    detector.photon.array = apply_psf(
        array=detector.photon.array, psf=psf, normalize_kernel=normalize_kernel
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import typing as t

import numpy as np
import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.charge_measurement import output_node_linearity_poly


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )


@pytest.mark.parametrize(
    "coefficients",
    [
        pytest.param([0, 1, 0.9]),
        pytest.param([5, 0.5, 0.9, 0.8]),
        pytest.param([3]),
        pytest.param([0, 3]),
    ],
)
def test_non_linearity_valid(ccd_5x5: CCD, coefficients: t.Sequence):
    """Test model 'non_linearity' with valid inputs."""
    detector = ccd_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    output_node_linearity_poly(detector=detector, coefficients=coefficients)


@pytest.mark.parametrize(
    "coefficients, exp_exc, exp_error",
    [
        pytest.param(
            [], ValueError, "Length of coefficient list should be more than 0."
        ),
        pytest.param(
            [0, -1],
            ValueError,
            "Signal array contains negative values after applying non-linearity model!",
        ),
    ],
)
def test_non_linearity_invalid(
    ccd_5x5: CCD, coefficients: t.Sequence, exp_exc, exp_error
):
    """Test model 'non_linearity' with valid inputs."""
    detector = ccd_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    with pytest.raises(exp_exc, match=exp_error):
        output_node_linearity_poly(detector=detector, coefficients=coefficients)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import typing as t

import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.readout_electronics import sar_adc


@pytest.fixture
def ccd_10x3() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(
            adc_bit_resolution=16, adc_voltage_range=(0.0, 10.0)
        ),
    )


def test_sar_adc(ccd_10x3: CCD):
    """Test model 'sar_adc' with valid inputs."""
    sar_adc(detector=ccd_10x3)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Tests for charge injection models."""

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDCharacteristics,
    CCDGeometry,
    CMOSCharacteristics,
    CMOSGeometry,
    Environment,
)
from pyxel.models.charge_generation import charge_blocks


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )


def test_charge_blocks_valid(ccd_10x10: CCD):
    """Test model 'charge_blocks' with valid inputs."""

    detector = ccd_10x10
    output = np.array(
        [
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    charge_blocks(detector=detector, charge_level=1, block_start=0, block_end=5)

    np.testing.assert_array_almost_equal(output, detector.charge.array)


@pytest.mark.parametrize(
    "charge_level, block_start, block_end, exp_exc, exp_error",
    [
        pytest.param(
            1,
            0,
            15,
            ValueError,
            "Block end not in range of the detector shape.",
        ),
        pytest.param(
            1,
            -4,
            5,
            ValueError,
            "Block start not in range of the detector shape.",
        ),
        pytest.param(
            -1,
            0,
            5,
            ValueError,
            "Charge level value should not be a negative number.",
        ),
    ],
)
def test_charge_blocks_inputs(
    ccd_10x10: CCD,
    charge_level: int,
    block_start: int,
    block_end: int,
    exp_exc,
    exp_error,
):
    """Test model 'charge_blocks' with bad inputs."""
    with pytest.raises(exp_exc, match=exp_error):
        charge_blocks(
            detector=ccd_10x10,
            charge_level=charge_level,
            block_start=block_start,
            block_end=block_end,
        )


def test_charge_blocks_with_cmos():
    """Test model 'charge_blocks' with a `CMOS` detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=4,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CMOSCharacteristics(),
    )

    with pytest.raises(TypeError, match="Expecting a CCD object for detector."):
        charge_blocks(detector=detector, charge_level=1)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import typing as t
from pathlib import Path

import numpy as np
import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.charge_generation import (
    conversion_with_qe_map,
    load_charge,
    simple_conversion,
)


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )
    return detector


@pytest.fixture
def valid_qe_map_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""
    data_2d = (
        np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
            ]
        )
        * 0.5
    )

    final_path = f"{tmp_path}/qe.npy"
    np.save(final_path, arr=data_2d)

    return final_path


@pytest.fixture
def invalid_qe_map_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""

    data_2d = np.array(
        [
            [1.0, 1.0, 0.5, 1.0, 1.0],
            [1.0, 2.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 2.0, 1.0],
            [0.5, 1.0, 1.0, 1.0, 1.0],
        ]
    )

    final_path = f"{tmp_path}/qe.npy"
    np.save(final_path, arr=data_2d)

    return final_path


@pytest.mark.parametrize(
    "qe",
    [
        pytest.param(0.5, id="valid"),
        pytest.param(None, id="valid_none"),
    ],
)
def test_simple_conversion_valid(ccd_5x5: CCD, qe: float):

    detector = ccd_5x5
    detector.characteristics.quantum_efficiency = 0.5

    array = np.ones((5, 5))
    detector.photon.array = array
    target = array * 0.5

    simple_conversion(detector=detector, quantum_efficiency=qe)

    np.testing.assert_array_almost_equal(detector.charge.array, target)


@pytest.mark.parametrize(
    "qe, exp_exc, exp_error",
    [
        pytest.param(1.5, ValueError, "Quantum efficiency not between 0 and 1."),
    ],
)
def test_simple_conversion_valid(
    ccd_5x5: CCD,
    qe: float,
    exp_exc,
    exp_error,
):

    with pytest.raises(exp_exc, match=exp_error):
        simple_conversion(detector=ccd_5x5, quantum_efficiency=qe)


def test_conversion_with_qe_valid(ccd_5x5: CCD, valid_qe_map_path: t.Union[str, Path]):

    detector = ccd_5x5

    conversion_with_qe_map(detector=detector, filename=valid_qe_map_path)


def test_simple_conversion_invalid(
    ccd_5x5: CCD, invalid_qe_map_path: t.Union[str, Path]
):

    with pytest.raises(
        ValueError, match="Quantum efficiency values not between 0 and 1."
    ):
        conversion_with_qe_map(detector=ccd_5x5, filename=invalid_qe_map_path)



import typing as t
from pathlib import Path

import numpy as np
import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment
from pyxel.models.charge_collection import fix_pattern_noise


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )
    return detector


@pytest.fixture
def valid_noise_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""
    data_2d = (
        np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
            ]
        )
        * 0.5
    )

    final_path = f"{tmp_path}/noise.npy"
    np.save(final_path, arr=data_2d)

    return final_path


def test_fix_pattern_noise_valid(ccd_5x5: CCD, valid_noise_path: t.Union[str, Path]):
    """Test function fix_pattern_noise with valid inputs."""

    detector = ccd_5x5

    array = np.ones((5, 5))
    detector.pixel.array = array
    target = array * 0.5

    fix_pattern_noise(detector=detector, filename=valid_noise_path)

    np.testing.assert_array_almost_equal(detector.pixel.array, target)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np

from pyxel.models.readout_electronics.amplifier_crosstalk import get_matrix


def test_get_matrix_with_list():
    """Test function 'get_matrix'."""
    matrix = [[1, 0.5], [0.5, 1], [0, 0]]
    data = get_matrix(matrix)

    exp_matrix = np.array([[1, 0.5], [0.5, 1], [0, 0]])
    np.testing.assert_equal(data, exp_matrix)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest

from pyxel.detectors import CMOS, CMOSCharacteristics, CMOSGeometry, Environment
from pyxel.models.charge_measurement import nghxrg


@pytest.fixture
def cmos_10x10() -> CMOS:
    """Create a valid CCD detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CMOSCharacteristics(),
    )

    # Generate fake pixels
    rs = np.random.RandomState(12345)
    detector.pixel.array = rs.normal(loc=10000, scale=500, size=(10, 10))

    return detector


@pytest.fixture
def cmos_10x15() -> CMOS:
    """Create a valid CCD detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=10,
            col=15,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CMOSCharacteristics(),
    )

    # Generate fake pixels
    rs = np.random.RandomState(12345)
    detector.pixel.array = rs.normal(loc=10000, scale=500, size=(10, 15))

    return detector


def test_nghxrg_10x10(cmos_10x10: CMOS):
    """Test model 'nghxrg' without parameters."""

    noise = [
        {"ktc_bias_noise": {"ktc_noise": 1, "bias_offset": 2, "bias_amp": 2}},
        {"white_read_noise": {"rd_noise": 1, "ref_pixel_noise_ratio": 2}},
        {"corr_pink_noise": {"c_pink": 1.0}},
        {"uncorr_pink_noise": {"u_pink": 1.0}},
        {"acn_noise": {"acn": 1.0}},
        {"pca_zero_noise": {"pca0_amp": 1.0}},
    ]

    nghxrg(detector=cmos_10x10, noise=noise)


# TODO: This test fails because 'y' and 'x' are not in the right order
def test_nghxrg_10x15(cmos_10x15: CMOS):
    """Test model 'nghxrg' without parameters."""

    noise = [
        {"ktc_bias_noise": {"ktc_noise": 1, "bias_offset": 2, "bias_amp": 2}},
        {"white_read_noise": {"rd_noise": 1, "ref_pixel_noise_ratio": 2}},
        {"corr_pink_noise": {"c_pink": 1.0}},
        {"uncorr_pink_noise": {"u_pink": 1.0}},
        {"acn_noise": {"acn": 1.0}},
        {"pca_zero_noise": {"pca0_amp": 1.0}},
    ]

    nghxrg(detector=cmos_10x15, noise=noise)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy

import numpy as np
import pandas as pd
import pytest

from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Detector, Environment


@pytest.fixture
def valid_ccd() -> CCD:
    """Create a valid `CCD`."""
    return CCD(
        geometry=CCDGeometry(
            row=100,
            col=120,
            total_thickness=123.1,
            pixel_horz_size=12.4,
            pixel_vert_size=34.5,
        ),
        environment=Environment(temperature=100.1),
        characteristics=CCDCharacteristics(
            quantum_efficiency=0.1,
            charge_to_volt_conversion=0.2,
            pre_amplification=3.3,
            full_well_capacity=10,
        ),
    )


@pytest.mark.parametrize(
    "other_obj, is_equal",
    [
        pytest.param(None, False, id="None"),
        pytest.param(
            CCD(
                geometry=CCDGeometry(row=100, col=120),
                environment=Environment(),
                characteristics=CCDCharacteristics(),
            ),
            False,
            id="Empty 'CCD'",
        ),
        pytest.param(
            CCD(
                geometry=CCDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(),
                characteristics=CCDCharacteristics(),
            ),
            False,
            id="Almost same parameters, different class",
        ),
        pytest.param(
            CCD(
                geometry=CCDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(temperature=100.1),
                characteristics=CCDCharacteristics(
                    quantum_efficiency=0.1,
                    charge_to_volt_conversion=0.2,
                    pre_amplification=3.3,
                    full_well_capacity=10,
                ),
            ),
            True,
            id="Same parameters, same class",
        ),
    ],
)
def test_is_equal(valid_ccd: CCD, other_obj, is_equal):
    """Test equality statement for `CCD`."""

    if is_equal:
        assert valid_ccd == other_obj
    else:
        assert valid_ccd != other_obj


def test_is_equal_with_arrays(valid_ccd: CCD):
    other_detector = deepcopy(valid_ccd)

    # Check that the deepcopy was applied
    assert valid_ccd.geometry is not other_detector.geometry
    assert valid_ccd.environment is not other_detector.environment
    assert valid_ccd.characteristics is not other_detector.characteristics
    assert valid_ccd._photon is not other_detector._photon
    assert valid_ccd._charge is not other_detector._charge
    assert valid_ccd._pixel is not other_detector._pixel
    assert valid_ccd._signal is not other_detector._signal
    assert valid_ccd._image is not other_detector._image

    # Generate random data
    shape = valid_ccd.geometry.row, valid_ccd.geometry.col  # type: t.Tuple[int, int]
    photon_2d = np.random.random(size=shape)
    pixel_2d = np.random.random(size=shape)
    signal_2d = np.random.random(size=shape)
    image_2d = np.random.random(size=shape)
    charge_2d = np.random.random(size=shape)

    # Apply the random data to 'valid_ccd' and 'other_detector'
    valid_ccd.photon.array = photon_2d.copy()
    valid_ccd.pixel.array = pixel_2d.copy()
    valid_ccd.signal.array = signal_2d.copy()
    valid_ccd.image.array = image_2d.copy()

    valid_ccd.charge._array = charge_2d.copy()
    valid_ccd.charge._frame = valid_ccd.charge._frame.append(
        {"charge": 1}, ignore_index=True
    )

    other_detector.photon.array = photon_2d.copy()
    other_detector.pixel.array = pixel_2d.copy()
    other_detector.signal.array = signal_2d.copy()
    other_detector.image.array = image_2d.copy()

    other_detector.charge._array = charge_2d.copy()
    other_detector.charge._frame = other_detector.charge._frame.append(
        {"charge": 1}, ignore_index=True
    )

    assert valid_ccd == other_detector


def comparison(dct, other_dct):
    assert set(dct) == set(other_dct) == {"version", "type", "data", "properties"}
    assert dct["version"] == other_dct["version"]
    assert dct["type"] == other_dct["type"]
    assert dct["properties"] == other_dct["properties"]

    assert (
        set(dct["data"])
        == set(other_dct["data"])
        == {"photon", "pixel", "signal", "image", "charge"}
    )
    np.testing.assert_equal(dct["data"]["photon"], other_dct["data"]["photon"])
    np.testing.assert_equal(dct["data"]["pixel"], other_dct["data"]["pixel"])
    np.testing.assert_equal(dct["data"]["signal"], other_dct["data"]["signal"])
    np.testing.assert_equal(dct["data"]["image"], other_dct["data"]["image"])

    assert (
        set(dct["data"]["charge"])
        == set(other_dct["data"]["charge"])
        == {"array", "frame"}
    )
    np.testing.assert_equal(
        dct["data"]["charge"]["array"], other_dct["data"]["charge"]["array"]
    )
    pd.testing.assert_frame_equal(
        dct["data"]["charge"]["frame"], other_dct["data"]["charge"]["frame"]
    )


@pytest.mark.parametrize("klass", [CCD, Detector])
@pytest.mark.parametrize(
    "obj, exp_dict",
    [
        pytest.param(
            CCD(
                geometry=CCDGeometry(row=100, col=120),
                environment=Environment(),
                characteristics=CCDCharacteristics(),
            ),
            {
                "version": 1,
                "type": "CCD",
                "properties": {
                    "geometry": {
                        "row": 100,
                        "col": 120,
                        "total_thickness": None,
                        "pixel_horz_size": None,
                        "pixel_vert_size": None,
                    },
                    "environment": {"temperature": None},
                    "characteristics": {
                        "quantum_efficiency": None,
                        "charge_to_volt_conversion": None,
                        "pre_amplification": None,
                        "full_well_capacity": None,
                        "adc_bit_resolution": None,
                        "adc_voltage_range": None,
                    },
                },
                "data": {
                    "photon": np.zeros(shape=(100, 120)),
                    "pixel": np.zeros(shape=(100, 120)),
                    "signal": np.zeros(shape=(100, 120)),
                    "image": np.zeros(shape=(100, 120)),
                    "charge": {
                        "array": np.zeros(shape=(100, 120)),
                        "frame": pd.DataFrame(
                            columns=[
                                "charge",
                                "number",
                                "init_energy",
                                "energy",
                                "init_pos_ver",
                                "init_pos_hor",
                                "init_pos_z",
                                "position_ver",
                                "position_hor",
                                "position_z",
                                "velocity_ver",
                                "velocity_hor",
                                "velocity_z",
                            ],
                            dtype=float,
                        ),
                    },
                },
            },
            id="Default parameters",
        ),
        pytest.param(
            CCD(
                geometry=CCDGeometry(
                    row=100,
                    col=120,
                    total_thickness=123.1,
                    pixel_horz_size=12.4,
                    pixel_vert_size=34.5,
                ),
                environment=Environment(temperature=100.1),
                characteristics=CCDCharacteristics(
                    quantum_efficiency=0.1,
                    charge_to_volt_conversion=0.2,
                    pre_amplification=3.3,
                    full_well_capacity=10,
                    adc_bit_resolution=16,
                    adc_voltage_range=(0.0, 10.0),
                ),
            ),
            {
                "version": 1,
                "type": "CCD",
                "properties": {
                    "geometry": {
                        "row": 100,
                        "col": 120,
                        "total_thickness": 123.1,
                        "pixel_horz_size": 12.4,
                        "pixel_vert_size": 34.5,
                    },
                    "environment": {"temperature": 100.1},
                    "characteristics": {
                        "quantum_efficiency": 0.1,
                        "charge_to_volt_conversion": 0.2,
                        "pre_amplification": 3.3,
                        "full_well_capacity": 10,
                        "adc_bit_resolution": 16,
                        "adc_voltage_range": (0.0, 10.0),
                    },
                },
                "data": {
                    "photon": np.zeros(shape=(100, 120)),
                    "pixel": np.zeros(shape=(100, 120)),
                    "signal": np.zeros(shape=(100, 120)),
                    "image": np.zeros(shape=(100, 120)),
                    "charge": {
                        "array": np.zeros(shape=(100, 120)),
                        "frame": pd.DataFrame(
                            columns=[
                                "charge",
                                "number",
                                "init_energy",
                                "energy",
                                "init_pos_ver",
                                "init_pos_hor",
                                "init_pos_z",
                                "position_ver",
                                "position_hor",
                                "position_z",
                                "velocity_ver",
                                "velocity_hor",
                                "velocity_z",
                            ],
                            dtype=float,
                        ),
                    },
                },
            },
            id="CCD fully defined",
        ),
    ],
)
def test_to_and_from_dict(klass, obj, exp_dict):
    """Test methods 'to_dict', 'from_dict'."""
    assert type(obj) == CCD

    # Convert from `CCD` to a `dict`
    dct = obj.to_dict()
    comparison(dct, exp_dict)

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    comparison(copied_dct, exp_dict)

    # Convert from `dict` to `CCD`
    other_obj = klass.from_dict(copied_dct)
    assert type(other_obj) == CCD
    assert obj == other_obj
    assert obj is not other_obj
    comparison(copied_dct, exp_dict)


@pytest.mark.parametrize("klass", [CCD, Detector])
def test_to_and_from_dict_with_arrays_no_frame(valid_ccd: CCD, klass):
    # Generate random data
    shape = valid_ccd.geometry.row, valid_ccd.geometry.col  # type: t.Tuple[int, int]
    photon_2d = np.random.random(size=shape)
    pixel_2d = np.random.random(size=shape)
    signal_2d = np.random.random(size=shape)
    image_2d = np.random.random(size=shape)
    charge_2d = np.random.random(size=shape)

    valid_ccd.photon.array = photon_2d.copy()
    valid_ccd.pixel.array = pixel_2d.copy()
    valid_ccd.signal.array = signal_2d.copy()
    valid_ccd.image.array = image_2d.copy()

    valid_ccd.charge._array = charge_2d.copy()
    # valid_ccd.charge._frame=valid_ccd.charge._frame.append({"charge": 1}, ignore_index=True)

    # Convert from the detector to a `dict`
    dct = valid_ccd.to_dict()

    exp_dict = {
        "version": 1,
        "type": "CCD",
        "properties": {
            "geometry": {
                "row": 100,
                "col": 120,
                "total_thickness": 123.1,
                "pixel_horz_size": 12.4,
                "pixel_vert_size": 34.5,
            },
            "environment": {"temperature": 100.1},
            "characteristics": {
                "quantum_efficiency": 0.1,
                "charge_to_volt_conversion": 0.2,
                "pre_amplification": 3.3,
                "full_well_capacity": 10,
                "adc_bit_resolution": None,
                "adc_voltage_range": None,
            },
        },
        "data": {
            "photon": photon_2d.copy(),
            "pixel": pixel_2d.copy(),
            "signal": signal_2d.copy(),
            "image": image_2d.copy(),
            "charge": {
                "array": charge_2d.copy(),
                "frame": pd.DataFrame(
                    columns=[
                        "charge",
                        "number",
                        "init_energy",
                        "energy",
                        "init_pos_ver",
                        "init_pos_hor",
                        "init_pos_z",
                        "position_ver",
                        "position_hor",
                        "position_z",
                        "velocity_ver",
                        "velocity_hor",
                        "velocity_z",
                    ],
                    dtype=float,
                ),
            },
        },
    }

    comparison(dct, exp_dict)

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    comparison(copied_dct, exp_dict)

    # Convert from `dict` to `CCD`
    new_detector = klass.from_dict(copied_dct)
    assert type(new_detector) == CCD
    assert valid_ccd == new_detector
    assert valid_ccd is not new_detector
    comparison(copied_dct, exp_dict)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


import pytest

from pyxel.detectors import (
    APD,
    CCD,
    CMOS,
    APDCharacteristics,
    APDGeometry,
    CCDCharacteristics,
    CCDGeometry,
    CMOSCharacteristics,
    CMOSGeometry,
    Environment,
)
from pyxel.models.charge_measurement import (
    output_node_noise,
    output_node_noise_cmos,
    readout_noise_saphira,
)


@pytest.fixture
def ccd_5x10() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=5,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )


@pytest.fixture
def cmos_5x10() -> CMOS:
    """Create a valid CMOS detector."""
    return CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CMOSCharacteristics(charge_to_volt_conversion=1.0e-6),
    )


@pytest.fixture
def apd_5x5() -> APD:
    """Create a valid CCD detector."""
    return APD(
        geometry=APDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=APDCharacteristics(
            quantum_efficiency=1.0,
            adc_voltage_range=(0.0, 10.0),
            adc_bit_resolution=16,
            full_well_capacity=100000,
            avalanche_gain=1.0,
            pixel_reset_voltage=5.0,
            roic_gain=0.8,
        ),
    )


@pytest.mark.parametrize("detector_type", ["ccd", "cmos"])
@pytest.mark.parametrize(
    "std_deviation",
    [pytest.param(1.0, id="std positive"), pytest.param(0.0, id="std null")],
)
def test_output_node_noise(
    ccd_5x10: CCD,
    cmos_5x10: CMOS,
    detector_type: str,
    std_deviation: float,
):
    """Test model 'output_node_noise' with valid inputs."""
    if detector_type == "ccd":
        detector = ccd_5x10
    elif detector_type == "cmos":
        detector = cmos_5x10
    else:
        raise NotImplementedError

    output_node_noise(detector=detector, std_deviation=std_deviation)


@pytest.mark.parametrize("detector_type", ["ccd", "cmos"])
def test_output_node_noise_bad_std(ccd_5x10: CCD, cmos_5x10: CMOS, detector_type: str):
    """Test model 'output_node_noise' with invalid input(s)."""
    if detector_type == "ccd":
        detector = ccd_5x10
    elif detector_type == "cmos":
        detector = cmos_5x10
    else:
        raise NotImplementedError

    with pytest.raises(ValueError, match="'std_deviation' must be positive."):
        output_node_noise(detector=detector, std_deviation=-1.0)


def test_output_node_noise_cmos(cmos_5x10: CMOS):
    """Test model 'output_node_noise_cmos' with valid inputs."""
    detector = cmos_5x10

    output_node_noise_cmos(detector=detector, readout_noise=1.0, readout_noise_std=0.1)


def test_output_node_noise_with_ccd(ccd_5x10: CCD):
    """Test model 'output_node_noise_ccd' with a 'CCD'."""
    detector = ccd_5x10

    with pytest.raises(TypeError, match="Expecting a 'CMOS' detector object."):
        output_node_noise_cmos(
            detector=detector, readout_noise=1.0, readout_noise_std=2.0
        )


def test_output_node_noise_invalid_noise(cmos_5x10: CMOS):
    """Test model 'output_node_noise_ccd' with a 'CCD'."""
    detector = cmos_5x10

    with pytest.raises(ValueError, match="'readout_noise_std' must be positive."):
        output_node_noise_cmos(
            detector=detector, readout_noise=1.0, readout_noise_std=-1.0
        )


def test_readout_noise_saphira_with_ccd(ccd_5x10: CCD):
    """Test model 'readout_noise_saphira' with a 'CCD'."""
    detector = ccd_5x10

    with pytest.raises(TypeError, match="Expecting a 'APD' detector object."):
        readout_noise_saphira(
            detector=detector, roic_readout_noise=0.1, controller_noise=0.1
        )


def test_readout_noise_saphira(apd_5x5: APD):
    """Test model 'readout_noise_saphira' with valid inputs."""
    detector = apd_5x5

    readout_noise_saphira(
        detector=detector, roic_readout_noise=0.1, controller_noise=0.1
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import typing as t

import pytest

from pyxel.calibration.util import check_ranges


@pytest.mark.parametrize(
    "target_range, out_range, rows, cols, readout_times",
    [
        # Nothing to check
        ([], [], 10, 10, None),
        ([0, 5, 0, 10], [], 5, 10, None),  # No 'readout_times'
        # 2D : (start_row, stop_row, start_col, stop_col)
        ([0, 5, 0, 10], [0, 5, 0, 10], 5, 10, None),  # No 'readout_times'
        ([0, 10, 0, 5], [0, 10, 0, 5], 10, 5, None),  # No 'readout_times'
        ([0, 5, 0, 10], [10, 15, 10, 20], 5, 10, None),  # No 'readout_times'
        ([0, 5, 0, 10], [0, 5, 0, 10], 5, 10, -1),  # Lowest 'readout_times'
        ([0, 5, 0, 10], [0, 5, 0, 10], 5, 10, 100),  # Highest 'readout_times'
        # 3D : (start_time, stop_time, start_row, stop_row, start_col, stop_col)
        (
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 20],
            10,
            20,
            5,
        ),  # No 'readout_times'
        (
            [0, 5, 0, 20, 0, 10],
            [0, 5, 0, 20, 0, 10],
            20,
            10,
            5,
        ),  # No 'readout_times'
        (
            [0, 5, 0, 10, 0, 20],
            [0, 5, 30, 40, 50, 70],
            10,
            20,
            5,
        ),  # No 'readout_times'
        (
            [0, 5, 20, 40, 0, 10],
            [0, 5, 0, 20, 0, 10],
            40,
            10,
            5,
        ),  # No 'readout_times'
    ],
)
def test_check_range_valid(
    target_range: list,
    out_range: list,
    rows: int,
    cols: int,
    readout_times: t.Optional[int],
):
    """Test valid values for function 'check_range'."""
    check_ranges(
        target_fit_range=target_range,
        out_fit_range=out_range,
        rows=rows,
        cols=cols,
        readout_times=readout_times,
    )


@pytest.mark.parametrize(
    "target_range, out_range, rows, cols, readout_times, exp_error",
    [
        # 'target_range' is neither 2D nor 3D
        pytest.param([0], [0, 5, 0, 10], 5, 10, None, "", id="Target 1 element"),
        pytest.param([0, 5], [0, 5, 0, 10], 5, 10, None, "", id="Target 2 elements"),
        pytest.param([0, 5, 0], [0, 5, 0, 10], 5, 10, None, "", id="Target 3 elements"),
        pytest.param(
            [0, 5, 10, 0, 5], [0, 5, 0, 10], 5, 10, None, "", id="Target 5 elements"
        ),
        # 'out_range' is neither 2D nor 3D
        pytest.param([0, 5, 0, 10], [0], 5, 10, None, "", id="Out 1 element"),
        pytest.param([0, 5, 0, 10], [0, 5], 5, 10, None, "", id="Out 2 element"),
        pytest.param([0, 5, 0, 10], [0, 5, 0], 5, 10, None, "", id="Out 3 element"),
        pytest.param(
            [0, 5, 0, 10], [0, 5, 0, 10, 11], 5, 10, None, "", id="Out 5 element"
        ),
        # 2D : (start_row, stop_row, start_col, stop_col)
        # Different span for 'target_range' and 'out_range'
        pytest.param(
            [0, 5, 0, 10],
            [0, 6, 0, 10],
            5,
            10,
            None,
            "Fitting ranges have different lengths in 1st dimension",
            id="1D length - too long",
        ),
        pytest.param(
            [0, 5, 0, 10],
            [1, 5, 0, 10],
            5,
            10,
            None,
            "Fitting ranges have different lengths in 1st dimension",
            id="1D length - too short",
        ),
        pytest.param(
            [0, 5, 0, 10],
            [0, 5, 0, 11],
            5,
            10,
            None,
            "Fitting ranges have different lengths in 2nd dimension",
            id="2D length - too long",
        ),
        pytest.param(
            [0, 5, 0, 10],
            [0, 5, 0, 9],
            5,
            10,
            None,
            "Fitting ranges have different lengths in 2nd dimension",
            id="2D length - too short",
        ),
        # 3D : (start_time, stop_time, start_row, stop_row, start_col, stop_col)
        pytest.param(
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 19],
            5,
            10,
            5,
            "Fitting ranges have different lengths in third dimension",
            id="3D length - too short",
        ),
        pytest.param(
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 21],
            5,
            10,
            5,
            "Fitting ranges have different lengths in third dimension",
            id="3D length - too long",
        ),
        pytest.param(
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 20],
            9,  # Too low
            20,
            20,
            "Value of target fit range is wrong",
            id="3D length - wrong target row range1",
        ),
        pytest.param(
            [0, 5, 30, 40, 0, 20],
            [0, 5, 30, 40, 0, 20],
            29,  # too low
            10,
            20,
            "Value of target fit range is wrong",
            id="3D length - wrong target row range2",
        ),
        pytest.param(
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 20],
            10,
            19,
            5,
            "Value of target fit range is wrong",
            id="3D length - wrong target col range1",
        ),
        pytest.param(
            [0, 5, 0, 10, 20, 40],
            [0, 5, 0, 10, 0, 20],
            10,
            39,  # too low
            5,
            "Value of target fit range is wrong",
            id="3D length - wrong target col range2",
        ),
        pytest.param(
            [0, 5, 0, 10, 20, 40],
            [0, 5, 0, 10, 0, 20],
            10,
            19,  # too low
            5,
            "Value of target fit range is wrong",
            id="3D length - wrong target col range3",
        ),
        pytest.param(
            [0, 5, 0, 10, 0, 20],
            [0, 5, 0, 10, 0, 20],
            10,
            20,
            None,
            "Target data is not a 3 dimensional array",
            id="3D length - Missing 'readout_times'",
        ),
        pytest.param(
            [10, 15, 0, 10, 0, 20],
            [10, 15, 0, 10, 0, 20],
            10,
            20,
            14,  # too low
            "Value of target fit range is wrong",
            id="3D length - wrong target time range1",
        ),
        pytest.param(
            [10, 15, 0, 10, 0, 20],
            [10, 15, 0, 10, 0, 20],
            10,
            20,
            9,  # too low
            "Value of target fit range is wrong",
            id="3D length - wrong target time range2",
        ),
    ],
)
def test_check_ranges_invalid(
    target_range: list,
    out_range: list,
    rows: int,
    cols: int,
    readout_times: t.Optional[int],
    exp_error,
):
    """Test valid values for function 'check_range'."""
    with pytest.raises(ValueError, match=exp_error):
        check_ranges(
            target_fit_range=target_range,
            out_fit_range=out_range,
            rows=rows,
            cols=cols,
            readout_times=readout_times,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import numpy as np
import pytest

from pyxel.detectors import (
    APD,
    CCD,
    APDCharacteristics,
    APDGeometry,
    CCDCharacteristics,
    CCDGeometry,
    Environment,
)
from pyxel.models.charge_generation import apd_gain


@pytest.fixture
def apd_5x5() -> APD:
    """Create a valid CCD detector."""
    return APD(
        geometry=APDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=APDCharacteristics(
            quantum_efficiency=1.0,
            adc_voltage_range=(0.0, 10.0),
            adc_bit_resolution=16,
            full_well_capacity=100000,
            avalanche_gain=2.0,
            pixel_reset_voltage=5.0,
            roic_gain=0.8,
        ),
    )


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(
            adc_voltage_range=(0.0, 10.0),
        ),
    )


def test_apd_gain_invalid_detector(ccd_5x5: CCD):
    """Test model 'apd_gain' with a 'CCD'."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a 'APD' detector object."):
        apd_gain(detector=detector)


def test_apd_gain(apd_5x5: APD):
    """Test model 'apd_gain' with a 'APD'."""
    detector = apd_5x5

    detector.charge.add_charge_array(array=np.ones((5, 5)))

    apd_gain(detector=detector)

    detector.charge.add_charge(
        particle_type="e",
        particles_per_cluster=np.array([1.0]),
        init_energy=np.array([0]),
        init_ver_position=np.array([0]),
        init_hor_position=np.array([0]),
        init_z_position=np.array([0]),
        init_ver_velocity=np.array([0]),
        init_hor_velocity=np.array([0]),
        init_z_velocity=np.array([0]),
    )

    assert detector.charge.frame_empty() is False

    apd_gain(detector=detector)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy

import pytest

from pyxel.detectors import CCDGeometry, Geometry


@pytest.mark.parametrize(
    "row, col, total_thickness, pixel_vert_size, pixel_horz_size",
    [(1, 1, 0.0, 0.0, 0.0), (10000, 10000, 10000.0, 1000.0, 1000.0)],
)
def test_create_valid_geometry(
    row, col, total_thickness, pixel_vert_size, pixel_horz_size
):
    """Test when creating a valid `Geometry` object."""
    _ = CCDGeometry(
        row=row,
        col=col,
        total_thickness=total_thickness,
        pixel_vert_size=pixel_vert_size,
        pixel_horz_size=pixel_horz_size,
    )


@pytest.mark.parametrize(
    "row, col, total_thickness, pixel_vert_size, pixel_horz_size, exp_exc",
    [
        pytest.param(0, 100, 100.0, 100.0, 100.0, ValueError, id="row == 0"),
        pytest.param(-1, 100, 100.0, 100.0, 100.0, ValueError, id="row < 0"),
        pytest.param(100, 0, 100.0, 100.0, 100.0, ValueError, id="col == 0"),
        pytest.param(100, -1, 100.0, 100.0, 100.0, ValueError, id="col < 0"),
        pytest.param(
            100, 100, -0.1, 100.0, 100.0, ValueError, id="total_thickness < 0."
        ),
        pytest.param(
            100, 100, 10000.1, 100.0, 100.0, ValueError, id="total_thickness > 10000."
        ),
    ],
)
def test_create_invalid_geometry(
    row, col, total_thickness, pixel_vert_size, pixel_horz_size, exp_exc
):
    """Test when creating an invalid `Geometry` object."""
    with pytest.raises(exp_exc):
        _ = CCDGeometry(
            row=row,
            col=col,
            total_thickness=total_thickness,
            pixel_vert_size=pixel_vert_size,
            pixel_horz_size=pixel_horz_size,
        )


@pytest.mark.parametrize(
    "other_obj, is_equal",
    [
        pytest.param(None, False, id="None"),
        pytest.param(CCDGeometry(row=100, col=120), False, id="Only two parameters"),
        pytest.param(
            CCDGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            True,
            id="Same parameters, same class",
        ),
        pytest.param(
            Geometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            False,
            id="Same parameters, different class",
        ),
    ],
)
def test_is_equal(other_obj, is_equal):
    """Test equality statement for CCDGeometry."""
    obj = CCDGeometry(
        row=100,
        col=120,
        total_thickness=123.1,
        pixel_horz_size=12.4,
        pixel_vert_size=34.5,
    )

    if is_equal:
        assert obj == other_obj
    else:
        assert obj != other_obj


@pytest.mark.parametrize(
    "obj, exp_dict",
    [
        (
            CCDGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            {
                "row": 100,
                "col": 120,
                "total_thickness": 123.1,
                "pixel_horz_size": 12.4,
                "pixel_vert_size": 34.5,
            },
        ),
    ],
)
def test_to_and_from_dict(obj, exp_dict):
    """Test methods 'to_dict', 'from_dict'."""
    assert type(obj) == CCDGeometry

    # Convert from `CCDGeometry` to a `dict`
    dct = obj.to_dict()
    assert dct == exp_dict

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    assert copied_dct == exp_dict

    # Convert from `dict` to `CCDGeometry`
    other_obj = CCDGeometry.from_dict(copied_dct)
    assert type(other_obj) == CCDGeometry
    assert obj == other_obj
    assert obj is not other_obj
    assert copied_dct == exp_dict



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import typing as t
from collections import abc
from pathlib import Path

import h5py as h5
import numpy as np
import pandas as pd

from pyxel import __version__

ATTRIBUTES = {
    "photon": {"name": "Photon", "unit": "photon/s"},
    "charge": {"name": "Charge", "unit": "electron"},
    "pixel": {"name": "Pixel", "unit": "electron"},
    "signal": {"name": "Signal", "unit": "volt"},
    "image": {"name": "Image", "unit": "adu"},
}  # type: t.Mapping[str, t.Mapping[str, str]]


def _store(
    h5file: h5.File,
    name: str,
    dct: t.Mapping[str, t.Union[int, float, pd.DataFrame, pd.Series, np.ndarray, dict]],
    attributes: t.Optional[t.Mapping[str, t.Mapping[str, str]]] = None,
) -> None:
    """Write data into a new HDF5 group.

    Parameters
    ----------
    h5file : h5.File
        Writeable HDF5 file object.
    name : str
        Name of the HDF5 group to create. (e.g. '/' or '/geometry')
    dct : dict
        Data to write into a HDF5 dataset.
    attributes : dict
        Attributes to store.
    """
    for key, value in dct.items():
        new_name = f"{name}/{key}"

        if isinstance(value, (int, float)) or value is None:
            if value is None:
                value = np.nan

            dataset = h5file.create_dataset(
                name=new_name, data=value, shape=()
            )  # type: h5.Dataset

            if attributes is not None and key in attributes:
                dataset.attrs.update(attributes[key])

        elif isinstance(value, pd.DataFrame):
            new_group = h5file.create_group(name=new_name)  # type: h5.Group
            new_group.attrs["type"] = "DataFrame"

            _store(h5file, name=new_name, dct=value.to_dict(orient="series"))

        elif isinstance(value, (pd.Series, np.ndarray, abc.Sequence)):
            dataset = h5file.create_dataset(name=new_name, data=np.asarray(value))

            if attributes is not None and key in attributes:
                dataset.attrs.update(attributes[key])

        elif isinstance(value, dict):
            new_group = h5file.create_group(name=new_name)

            if attributes is not None and key in attributes:
                new_group.attrs.update(attributes[key])

            _store(h5file, name=new_name, dct=value)

        else:
            raise NotImplementedError


def to_hdf5(filename: t.Union[str, Path], dct: t.Mapping[str, t.Any]) -> None:
    """Write data to a HDF5 file."""
    if dct["version"] != 1:
        raise NotImplementedError

    with h5.File(filename, mode="w") as h5file:
        # Write main attributes
        h5file.attrs["version"] = dct["version"]
        h5file.attrs["type"] = dct["type"]
        h5file.attrs["pyxel-version"] = str(__version__)

        _store(h5file, name="/geometry", dct=dct["properties"]["geometry"])
        _store(h5file, name="/environment", dct=dct["properties"]["environment"])
        _store(
            h5file, name="/characteristics", dct=dct["properties"]["characteristics"]
        )

        _store(h5file, name="/data", dct=dct["data"], attributes=ATTRIBUTES)


def _load(
    h5file: h5.File, name: str
) -> t.Union[None, int, float, t.Mapping[str, t.Any], np.ndarray, pd.DataFrame]:
    """Write data from a HDF5 group.

    Parameters
    ----------
    h5file : h5.File
        Readable HDF5 file object.
    name : str
        Name of the HDF5 group to read. (e.g. '/' or '/geometry')

    Returns
    -------
    dict
        Data to read from a HDF5 dataset.
    """
    dataset = h5file[name]  # type: t.Union[h5.Dataset, h5.Group]

    if isinstance(dataset, h5.Group):
        dct = {}
        for key, _ in h5file[name].items():
            result = _load(h5file, name=f"{name}/{key}")

            dct[key] = result

        if name.endswith("frame"):
            return pd.DataFrame.from_dict(dct)
        else:
            return dct

    elif isinstance(dataset, h5.Dataset):
        if dataset.ndim == 0:
            value = np.array(dataset)

            if np.isnan(value):
                return None
            elif np.issubdtype(dataset.dtype, np.integer):
                return int(value)
            elif np.issubdtype(dataset.dtype, np.floating):
                return float(value)
            else:
                raise TypeError
        else:
            return np.array(dataset)
    else:
        raise NotImplementedError


def from_hdf5(filename: t.Union[str, Path]) -> t.Mapping[str, t.Any]:
    """Read data from a HDF5 file."""
    dct = {}
    with h5.File(filename, mode="r") as h5file:
        # Read main attributes
        dct.update(h5file.attrs)

        # TODO: Use a JSON schema to validate 'dct'
        if "version" not in dct or "type" not in dct:
            raise ValueError("Missing 'version' and/or 'type' !")

        version = dct["version"]  # type: int
        if version != 1:
            raise NotImplementedError

        # Get properties
        properties = {}
        for name in ["geometry", "environment", "characteristics"]:  # type: str
            properties[name] = _load(h5file, name=f"/{name}")

        dct["properties"] = properties

        # Get data
        data = {}
        for name in h5file["/data"]:
            data[name] = _load(h5file, name=f"/data/{name}")

        dct["data"] = data

    return dct



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#

#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
import pytest

from pyxel.detectors import (
    CCD,
    CCDCharacteristics,
    CCDGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.photon_generation import stripe_pattern


@pytest.fixture
def ccd_20x20() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=20,
            col=20,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=CCDCharacteristics(),
    )
    detector._readout_properties = ReadoutProperties(num_steps=1)
    return detector


@pytest.mark.parametrize(
    "period, level, angle, startwith, time_scale",
    [
        pytest.param(
            2,
            10.0,
            0,
            0,
            1.0,
            id="valid",
        ),
        pytest.param(
            1,
            10.0,
            0,
            0,
            1.0,
            id="period_too_small",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
        pytest.param(
            5,
            10.0,
            0,
            0,
            1.0,
            id="period_not_multiple_of_2",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
    ],
)
def test_stripe_pattern(
    ccd_20x20: CCD,
    period: int,
    level: float,
    angle: int,
    startwith: int,
    time_scale: float,
):
    """Test input parameters for function 'stripe_pattern'."""

    stripe_pattern(
        detector=ccd_20x20,
        period=period,
        level=level,
        angle=angle,
        startwith=startwith,
        time_scale=time_scale,
    )
